/*
 ** farsta version 1.1 ** 13 May 1997 ** by Andrew Kresch
 **
 ** compute Gromov-Witten invariants for algebraic varieties using
 ** quantum associativity relations
 **
 ** stores numbers and linear relations
 **
 ** all artithmetic is with fixed precision rational numbers
 ** (precision set at compile time)
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include "fixed.h"
#define min(x,y) ((x) > (y) ? (y) : (x))
#define max(x,y) ((x) > (y) ? (x) : (y))
#define MAXNRELS 500
#define MAXLIN 600
#define MAXKNOWN 30000
#define MAXEXPR 5000
#define MAXRELEXPR 5000
#define MAXLINEXPR 10000
#define LBUFFER 50
/* pvect stuff */
#define MACAULAY 1
#define HUMAN 2
#define HUMANCOMPACT 3
#define HUMANFORREENTRY 4
#define CLEAR_DENOMS 1
#define COMMON_DENOMS 2
#define NO_DENOMS 3
#define NUMLEN(n) ((n) < 10 ? 1 : \
		    (n) < 100 ? 2 : \
		    (n) < 1000 ? 3 : \
		    (n) < 10000 ? 4 : \
		    (n) < 100000 ? 5 : 6 )

void fnicempvect(FILE *, qnumber *, int, int, int *,
char **, int *, int, int, int);
void nicempvect(qnumber *, int, int, int *, char **,
int *, int, int, int);
int fscanmvect(FILE *, qnumber *, int, int, int *,
char **, int *, int *, char **);
int scanmvect(qnumber *, int, int, int *, char **,
int *, int *, char **);
int fgetline(FILE *, char *, int);
int getline(char *, int);
void mgcd(znumber *, znumber *, znumber *);
int qmatinv(qnumber *, qnumber *, int *, int, int);

/* expression structure: linked list, length marker tells actual length,
e.g. 5N01 - N11 + N12, if we add N11 then get 5N01 pointing to N12 (only
two terms) but actual length is still 3, the 0 N11 is taking up dead space
in the middle -- call collapsee to reclaim this dead space */
/* if e is an expression then e[0] is useless except e[0].next points to
first term in the expression and e[0].d points to beginning of index data
storage; until things start cancelling, e[i].next points to e[i+1]
and e[i].d points to e[i-1].isize ints ahead of e[i-1].d */
typedef struct expr {
    qnumber coef;
    int isize;              /* size of indexing data */
    int *d;                 /* pointer to indexing data */
    struct expr *next;      /* must be AHEAD of us, or NULL for end of expr */
} expr;
typedef struct {
    char *name;
    int dim;            /* dim X */
    int rkHstar;        /* rank H^*X */
    char **cohname;     /* labels of cohomology basis elements */
			/* y0, (codimension 2 classes), (the rest) */
    int *namelen;       /* lengths (in characters) of labels */
    int rkH2;           /* rank H^2 */
    int *cohcodim;      /* codim of cohomology basis elts: 0 1 1..1 (#s >=2) */
    qnumber *potq;      /* matrix of d_ijk (class.pot.fcn) */
    qnumber *imatq;     /* intersection matrix */
		    /* ij entry is integral_X (e_i cup e_j), e_ basis elts */
			/* ij entry is stored in imat[i*rhHstar+j] */
    qnumber *imatinvq;  /* inverse of above */
    int rkPic;          /* rank of Picard group */
			/* by convention, cone of effective curves is */
			/* contained in region with all entries >= 0 */
    qnumber *cclassq;   /* curve class in the homology basis */
			/* [curve i]=sum (j=1..rkH2)cclass[i*rkH2+j-1]*e_j */
    int *cclass;
    int cccd;
    int *c1t;       /* integral_(c_i) c1(TX) (first Chern class of tg bndl)*/
    int *loc;           /* lower bound (vector) on the curve class */
    int nrels;          /* number of distinct relations, should be */
			/* m(m-1)(m^2-m+2), where m = rkHstar-1 */
    int *Ri;            /* instructions for producing each relation: */
    int *Rj;            /* coef of Tl^ of (Ti*Tj)*Tk = Ti*(Tj*Tk) */
    int *Rk;
    int *Rl;
    int *weight;        /* cohcodim[i]+cohcodim[j]+cohcodim[k]+cohcodim[l]-n */
} variety;

extern int ov;
int rkP;        /* rank of Picard */
int ni;         /* number of indices = rkPic + rkHstar - rkH2 - 1 */
int *vallist, *linlist;
qnumber value[MAXKNOWN];
#if 0
#define linrhs2 linrhs
expr *linrhs[MAXLIN], /* *linrhs2[MAXLIN], */ *linex, *linex2;
#else
expr *linrhs[MAXLIN], *linrhs2[MAXLIN], *linex, *linex2;
#endif
int nknown, nlinear;
int A;
int *expn, *expn2, *linexd, *linex2d, npterms;
qnumber *potvect;
int *restartbuf, restartable;
variety V;
int nsig;

int ztoi(znumber *z)
{
    if (z->len == 0)
	return 0;
    if (z->len != 1) {
	ov = 1;
	return 0;
    }
    return z->sign ? -z->s[0] : z->s[0];
}

void mysig(int sig)
{
    if (nsig++ == 10)
	signal(SIGINT, SIG_DFL);
}

void userinputsig(int sig)
{
    if (nsig++ == 10)
	signal(SIGINT, SIG_DFL);
    printf("Type control-D or Q to quit\n");
}

int killvariety(variety *v)
{
    int i;

    free(v->name);
    for (i=0; i<v->rkHstar; i++)
	free(v->cohname[i]);
    free(v->cohname);
    free(v->namelen);
    free(v->cohcodim);
    free(expn);
    free(expn2);
    free(v->potq);
    free(v->imatq);
    free(v->imatinvq);
    free(potvect);
    free(v->cclassq);
    free(v->cclass);
    free(v->c1t);
    free(v->loc);
    free(vallist);
    free(restartbuf);
    free(v->Ri);
    free(v->Rj);
    free(v->Rk);
    free(v->Rl);
    free(v->weight);
    if (linlist) {
	free(linlist);
	linlist = NULL;
	free(linex);
	free(linex2);
	free(linexd);
	free(linex2d);
    }
}

/* read in variety from user, return 1 for error, 0 normal */
int scanvariety(FILE *from, variety *v, char *name)
{
    int i, j, jj, k, s, rk, vv[3];
    int c, d, e, f, g;
    int *pow, *scr, *xxxx;
    char str[1000], *t, *ot;
    qnumber *xxx, q1;
    znumber one, x, y, z, denom;
    qnumber denomq, w;

    v->name = strdup(name);
    if (from == stdin)
	printf("dim(%s) = ?         >", name);
    if (fgetline(from, str, 999) == EOF)
	return 1;
    if (sscanf(str, "%d", &i) != 1)
	return 1;
    if (i<=1)
	return 1;
    v->dim = i;
    if (from == stdin)
	printf("rank H^*(%s) = ?    >", name);
    if (fgetline(from, str, 999) == EOF)
	return 1;
    if (sscanf(str, "%d", &rk) != 1)
	return 1;
    if (rk<=1)
	return 1;
    v->rkHstar = rk;
    if (from == stdin)
	printf("basis H^*(%s) = ?   >", name);
    v->cohname = (char **) malloc(rk*sizeof(char *));
    str[0] = '\0';
    t = str;
    for (i=0; i<rk; ) {
	if (t[0] == '\0') {
	    if (fgetline(from, str, 999) == EOF)
		return 1;
	    t = str;
	}
	while (*t != '\0' && isspace(*t))
	    t++;
	if (t[0] != '\0') {
	    ot = t;
	    while (*t != '\0' && !isspace(*t))
		t++;
	    if (*t != '\0') {
		*t = '\0';
		v->cohname[i++] = strdup(ot);
		t++;
	    } else {
		v->cohname[i++] = strdup(ot);
	    }
	}
    }
    v->namelen = (int *)malloc(rk*sizeof(int));
    for (i=0; i<rk; i++)
	v->namelen[i] = strlen(v->cohname[i]);
    if (from == stdin)
	printf("rank H^2(%s) = ?    >", name);
    if (fgetline(from, str, 999) == EOF)
	return 1;
    if (sscanf(str, "%d", &i) != 1)
	return 1;
    if (i<=0 || i > rk - 2)
	return 1;
    v->rkH2 = i;
    v->cohcodim = (int *)malloc(rk*sizeof(int));
    v->cohcodim[0] = 0;
    for (i=1; i<=v->rkH2; i++)
	v->cohcodim[i] = 1;
    for (i=v->rkH2+1; i<rk; i++) {
	if (from == stdin)
	    printf("     (complex) codim(%s) = ?  %*s>",
			v->cohname[i], 10-v->namelen[i], "");
	if (fgetline(from, str, 999) == EOF)
	    return 1;
	if (sscanf(str, "%d", &j) != 1)
	    return 1;
	if (j<=1 || j > v->dim)
	    return 1;
	v->cohcodim[i] = j;
    }
    if (from == stdin)
	printf("enter classical potential function:\n");


/* homogeneous polynomials in rk variables of degree 3 */
/* number of terms is (rk + 3 - 1) choose 3 */
    npterms = (rk+2)*(rk+1)*rk/6;
/* DEBUG printf("*****  rk  %d   npterms    %d\n", rk, npterms); */
    pow = (int *)malloc((rk - 1)*sizeof(int));
    potvect = (qnumber *)malloc(npterms*sizeof(qnumber));
    expn = (int *)malloc(npterms*rk*sizeof(int));
    for (j=0; j<rk-1; j++)
	pow[j] = expn[j] = 0;
    expn[rk-1] = 3;
    for (i=1; i<npterms; i++) {
	s = 0;
	for (j=0; j<rk-1; j++)
	    s += pow[j];
	j = rk-2;
	while (s >= 3) {
	    s -= pow[j];
	    pow[j] = 0;
	    j--;
	}
	pow[j]++;
	for (j=0; j<rk-1; j++)
	    expn[i*rk+j] = pow[j];
	expn[i*rk+rk-1] = 3 - 1 - s;
    }

    scr = (int *)malloc(max(npterms,rk*rk*rk)*sizeof(int));
/* DEBUG printf("*****  npterms    %d\n", npterms); */
    j = fscanmvect(from, potvect, npterms, rk, expn, v->cohname,
			v->namelen, scr, NULL);
    if (j == EOF)
	return 1;
    if (j) {
	return 1;
    }
    c = 0;
    for (i=0; i<npterms; i++) {
	if (qsgn(&potvect[i])) {
	    for (k=0; k<3; k++)
		vv[k] = 0;
	    k = 0;
	    for (j=0; j<rk; j++)
		for (jj=0; jj<expn[i*rk+j]; jj++)
		    vv[k++] = j;
	    if (v->cohcodim[vv[0]] + v->cohcodim[vv[1]] + v->cohcodim[vv[2]]
						    != v->dim) {
		if (c == 0) {
		    printf(
"error in potential function -- some terms have wrong degree\n");
		    c = 1;
		}
		nicempvect(&potvect[i], 1, v->rkHstar, &expn[i*rk],
			v->cohname, v->namelen, HUMAN, NO_DENOMS, 1);
	    }
	}
    }
    if (c)
	return 1;
    v->potq = (qnumber *)malloc(rk*rk*rk*sizeof(qnumber));
    for (i=0; i<rk; i++)
	for (j=0; j<rk; j++)
	    for (k=0; k<rk; k++)
		qzero(&v->potq[i*rk*rk+j*rk+k]);
    for (i=0; i<npterms; i++) {
	if (qsgn(&potvect[i])) {
	    for (k=0; k<3; k++)
		vv[k] = 0;
	    k = 0;
	    for (j=0; j<rk; j++)
		for (jj=0; jj<expn[i*rk+j]; jj++)
		    vv[k++] = j;
	    qadd(&v->potq[vv[0]*rk*rk+vv[1]*rk+vv[2]], &potvect[i], &q1);
			    qcopy(&q1, &v->potq[vv[0]*rk*rk+vv[1]*rk+vv[2]]);
	    qadd(&v->potq[vv[0]*rk*rk+vv[2]*rk+vv[1]], &potvect[i], &q1);
			    qcopy(&q1, &v->potq[vv[0]*rk*rk+vv[2]*rk+vv[1]]);
	    qadd(&v->potq[vv[1]*rk*rk+vv[0]*rk+vv[2]], &potvect[i], &q1);
			    qcopy(&q1, &v->potq[vv[1]*rk*rk+vv[0]*rk+vv[2]]);
	    qadd(&v->potq[vv[1]*rk*rk+vv[2]*rk+vv[0]], &potvect[i], &q1);
			    qcopy(&q1, &v->potq[vv[1]*rk*rk+vv[2]*rk+vv[0]]);
	    qadd(&v->potq[vv[2]*rk*rk+vv[0]*rk+vv[1]], &potvect[i], &q1);
			    qcopy(&q1, &v->potq[vv[2]*rk*rk+vv[0]*rk+vv[1]]);
	    qadd(&v->potq[vv[2]*rk*rk+vv[1]*rk+vv[0]], &potvect[i], &q1);
			    qcopy(&q1, &v->potq[vv[2]*rk*rk+vv[1]*rk+vv[0]]);
	}
    }
    v->imatq = (qnumber *)malloc(rk*rk*sizeof(qnumber));
    v->imatinvq = (qnumber *)malloc(rk*rk*sizeof(qnumber));
    xxx = (qnumber *)malloc((2*rk)*rk*sizeof(qnumber));
    xxxx = (int *)malloc(rk*sizeof(int));
    for (i=0; i<rk; i++)
	for (j=0; j<rk; j++)
	    qcopy(&v->potq[i*rk+j], &v->imatq[i*rk+j]);
#if 0
    for (i=0; i<npterms; i++) {
	if (qsgn(&potvect[i])) {
	    k = 0;
	    for (j=0; j<rk; j++)
		k += expn[i*rk+j];
	    if (k == 3 && expn[i*rk] > 0) {
		if (expn[i*rk] == 3) {
		    printf(
"paring y0.y0 = y0 not allowed on positive dimensional variety\n");
		    return 1;
		}
		if (expn[i*rk] == 2) {
		    c = 0;
		    for (j=1; j<rk; j++) {
			if (expn[i*rk+j] == 1) {
			    c = j;
			    break;
			}
		    }
		    if (c == 0) {
			printf("irrecoverable error\n");
			exit(1);
		    }
				/* term of the form y0^2 yi */
		    v->imat[0*rk+c] = (v->imat[0*rk+c] + 2*potvect[i])%PRIME;
		    v->imat[c*rk+0] = (v->imat[c*rk+0] + 2*potvect[i])%PRIME;
		} else {
		    c = 0;
		    for (j=1; j<rk; j++) {
			if (expn[i*rk+j] == 2) {
			    c = j;
			    break;
			}
		    }
		    if (c) {            /* term of the form y_0 y_i^2 */
			v->imat[c*rk+c] = (v->imat[c*rk+c]+2*potvect[i])%PRIME;
		    } else {            /* term of the form y_0 y_i y_j */
			c = d = 0;
			for (j=1; j<rk; j++) {
			    if (expn[i*rk+j] == 1) {
				c = j;
				break;
			    }
			}
			for (j=c+1; j<rk; j++) {
			    if (expn[i*rk+j] == 1) {
				d = j;
				break;
			    }
			}
			if (c == 0 || d == 0) {
			    printf("irrecoverable error\n");
			    exit(1);
			}
			v->imat[c*rk+d] = (v->imat[c*rk+d] + potvect[i])%PRIME;
			v->imat[d*rk+c] = (v->imat[d*rk+c] + potvect[i])%PRIME;
		    }
		}
	    }
	}
    }
#endif
/* put the inverse of iform in iiform */
    for (i=0; i<rk; i++)
	for (j=0; j<rk; j++)
	    qcopy(&v->imatq[i*rk+j], &v->imatinvq[i*rk+j]);
    i = qmatinv(v->imatinvq, xxx, xxxx, rk, rk);
    free(xxx);
    free(xxxx);
    if (i) {
	printf("error: intersection matrix is singular\n");
	return 1;
    }
    if (from == stdin)
	printf("rank Pic(%s) = ?    >", name);
    if (fgetline(from, str, 999) == EOF)
	return 1;
    if (sscanf(str, "%d", &i) != 1)
	return 1;
    if (i<=0)
	return 1;
    v->rkPic = i;
    expn2 = (int *)malloc((v->rkH2)*(v->rkH2)*sizeof(int));
    for (i=0; i<v->rkH2; i++)
	for (j=0; j<v->rkH2; j++)
	    expn2[i*v->rkH2+j] = i==j ? 1:0;
    v->cclassq = (qnumber *)malloc((v->rkPic)*(v->rkH2)*sizeof(qnumber));
    v->cclass = (int *)malloc((v->rkPic)*(v->rkH2)*sizeof(int));
    for (i=0; i<v->rkPic; i++) {
	if (from == stdin)
	    printf("      PD(B_%d) in H^2(%s) = ?           >", i+1, name);
	j = fscanmvect(from, &v->cclassq[i*v->rkH2], v->rkH2, v->rkH2,
			expn2, v->cohname + 1, v->namelen + 1, scr, NULL);
	if (j == EOF)
	    return 1;
	if (j) {
	    return 1;
	}
    }


    zset(&one, 1);
    zset(&denom, 1);
    for (i=0; i<(v->rkPic)*(v->rkH2); i++) {
	if (qsgn(&v->cclassq[i])) {
	    mgcd(&denom, &v->cclassq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->cclassq[i].d, &y, &denom);
	}
    }
    ztoq(&denom, &denomq);
    v->cccd = ztoi(&denom);
    for (i=0; i<(v->rkPic)*(v->rkH2); i++) {
	if (qsgn(&v->cclassq[i])) {
	    qqmul(&v->cclassq[i], &denomq, &w);
	    v->cclass[i] = ztoi(&w.n);
	} else {
	    v->cclass[i] = 0;
	}
    }



    if (from == stdin)
	printf("T%s = tangent bundle on %s\n", name, name);
    v->c1t = (int *)malloc(v->rkPic*sizeof(int));
    for (i=0; i<v->rkPic; i++) {
	if (from == stdin)
	    printf("      int_(B_%d) c_1(T%s) = ?           >", i+1, name);
	if (fgetline(from, str, 999) == EOF)
	    return 1;
	if (sscanf(str, "%d", &j) != 1)
	    return 1;
	v->c1t[i] = j;
    }

    v->loc = (int *)malloc(v->rkPic*sizeof(int));
    for (i=0; i<v->rkPic; i++)
	v->loc[i] = 0;

    rkP = v->rkPic;
    ni = v->rkPic + rk - v->rkH2 - 1;
    vallist = (int *)malloc(MAXKNOWN*ni*sizeof(int));
    nknown = 0;
    restartable = 0;
    free(pow);
    free(scr);
/* &etc */
    return 0;
}

void printvariety(FILE *where, variety *v, int form)
{
    int i, j, havedenom;
    char u1[100], u2[100];
    znumber one, x, y, z, denom;
    qnumber denomq, w;

    if (form) {
	fprintf(where, "A %s\n%d\n%d\n", v->name, v->dim, v->rkHstar);
	fprintf(where, "%s", v->cohname[0]);
	j = v->namelen[0];
	for (i=1; i<v->rkHstar; i++) {
	    j += v->namelen[i]+1;
	    if (j >= 70) {
		j = v->namelen[i];
		fprintf(where, "\n%s", v->cohname[i]);
	    } else {
		fprintf(where, " %s", v->cohname[i]);
	    }
	}
	putc('\n', where);
	fprintf(where, "%d\n", v->rkH2);
	for (i=v->rkH2+1; i<v->rkHstar; i++)
	    fprintf(where, "%d\n", v->cohcodim[i]);
    } else {
	fprintf(where, "Current variety is %s\n", v->name);
	fprintf(where,
"dim(%s) = %d      rk H^*(%s) = %d     rk H^2(%s) = %d     rk Pic(%s) = %d\n",
	    v->name, v->dim, v->name, v->rkHstar, v->name, v->rkH2,
						    v->name, v->rkPic);
	fprintf(where, "basis for H^*(%s)    codim\n", v->name);
	for (i=0; i<v->rkHstar; i++)
	    fprintf(where, "        %s  %*s%d\n", v->cohname[i],
				12-v->namelen[i], "", v->cohcodim[i]);
	fprintf(where, "classical potential function\n");
    }
    fnicempvect(where, potvect, npterms, v->rkHstar, expn, v->cohname,
		v->namelen, form ? HUMANFORREENTRY : HUMAN, NO_DENOMS, 1);
    if (form) {
	fprintf(where, "%d\n", v->rkPic);
	for (i=0; i<v->rkPic; i++)
	    fnicempvect(where, &v->cclassq[i*v->rkH2], v->rkH2, v->rkH2,
		    expn2, v->cohname + 1, v->namelen + 1, HUMANFORREENTRY,
		    NO_DENOMS, 0);
	for (i=0; i<v->rkPic; i++)
	    fprintf(where, "%d\n", v->c1t[i]);
    } else {
	fprintf(where, "yielding intersection form\n");
	zset(&one, 1);
	zset(&denom, 1);
	for (i=0; i<v->rkHstar*v->rkHstar; i++) {
	    mgcd(&denom, &v->imatq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->imatq[i].d, &y, &denom);
	}
	havedenom = zgle(&denom, &one) ? 1 : 0;
	ztoq(&denom, &denomq);
	if (havedenom) {
	    sprintf(u1, "  (1/%s)  (", ztodec(&denom));
	    strcpy(u2, u1);
	    for (i=0; i<strlen(u2)-1; i++)
		u2[i] = ' ';
	} else {
	    strcpy(u1, "          (");
	    strcpy(u2, u1);
	}
	for (i=0; i<v->rkHstar; i++) {
	    printf("%s", i==v->rkHstar/2 ? u1 : u2);
	    for (j=0; j<v->rkHstar; j++) {
		qqmul(&v->imatq[i*v->rkHstar+j], &denomq, &w);
		fprintf(where,
		    "%3s%s", qtodec(&w), j==v->rkHstar-1?"  )\n":" ");
	    }
	}
	fprintf(where, "with inverse\n");

	zset(&denom, 1);
	for (i=0; i<v->rkHstar*v->rkHstar; i++) {
	    mgcd(&denom, &v->imatinvq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->imatinvq[i].d, &y, &denom);
	}
	havedenom = zgle(&denom, &one) ? 1 : 0;
	ztoq(&denom, &denomq);
	if (havedenom) {
	    sprintf(u1, "  (1/%s)  (", ztodec(&denom));
	    strcpy(u2, u1);
	    for (i=0; i<strlen(u2)-1; i++)
		u2[i] = ' ';
	} else {
	    strcpy(u1, "          (");
	    strcpy(u2, u1);
	}
	for (i=0; i<v->rkHstar; i++) {
	    fprintf(where, "%s", i==v->rkHstar/2 ? u1 : u2);
	    for (j=0; j<v->rkHstar; j++) {
		qqmul(&v->imatinvq[i*v->rkHstar+j], &denomq, &w);
		fprintf(where,
		    "%3s%s", qtodec(&w), j==v->rkHstar-1?"  )\n":" ");
	    }
	}

	fprintf(where, "line bundles B_i satisfy\n");
	for (i=0; i<v->rkPic; i++) {
	    fprintf(where, "c_1(B_%d) = ", i+1);
	    fnicempvect(where, &v->cclassq[i*v->rkH2], v->rkH2, v->rkH2,
		    expn2, v->cohname + 1, v->namelen + 1, HUMANCOMPACT,
		    COMMON_DENOMS, 0);
	}
	for (i=0; i<v->rkPic; i++)
	    fprintf(where, "int_(B_%d) c_1(T%s) = %d\n", i+1, v->name,
							v->c1t[i]);
    }
}

void sort3(int *x)
{
    int j;

    if (x[1] > x[2]) {
	j = x[1];
	x[1] = x[2];
	x[2] = j;
    }
    if (x[0] > x[1]) {
	j = x[0];
	x[0] = x[1];
	x[1] = j;
    }
    if (x[1] > x[2]) {
	j = x[1];
	x[1] = x[2];
	x[2] = j;
    }
}

/* massage a term in the differential relations */
/* convert, e.g. 3 G_2,3,2 G_1,3,2 to 3 G_1,2,3 G_2,2,3 */
void massagedterm(expr *t)
{
    int j;
    int *x;

    x = t->d;
    if (t->isize == 3) {
	sort3(x);
	if (x[0] == 0)
	    qzero(&t->coef);
    } else if (t->isize == 6) {
	sort3(x);
	sort3(x+3);
	if (x[0] > x[3] ||
		(x[0] == x[3] && x[1] > x[4]) ||
		(x[0] == x[3] && x[1] == x[4] && x[2] > x[5])) {
	    j = x[0];
	    x[0] = x[3];
	    x[3] = j;
	    j = x[1];
	    x[1] = x[4];
	    x[4] = j;
	    j = x[2];
	    x[2] = x[5];
	    x[5] = j;
	}
	if (x[0] == 0)
	    qzero(&t->coef);
    }
}

/* massage a term in the enumerative relations */
void massagenterm(expr *t)
{
    int i, b;
    if (t->isize == 2*ni) {
	for (i=0; i<ni && t->d[i] == t->d[ni+i]; i++)
	    ;
	if (i<ni && t->d[i] > t->d[ni+i]) {
	    for (i=0; i<ni; i++) {
		b = t->d[i];
		t->d[i] = t->d[ni+i];
		t->d[ni+i] = b;
	    }
	}
    }
    if (t->isize > 0) {
	for (i=0; i<rkP && t->d[i] == 0; i++)       /* N( 0 ; any ) = 0 */
	    ;
	if (i == rkP)
	    qzero(&t->coef);
    }
}

/* return c = a choose b */
void cmb(int a, int b, znumber *c)
{
    int bb, i;
    znumber cc, u, v;

    if (a < 0) {
	zzero(c);
	return;
    }
    if (b < 0) {
	zzero(c);
	return;
    }
    if (b == 0) {
	zset(c, 1);
	return;
    }
    bb = a - b;
    if (bb < 0) {
	zzero(c);
	return;
    }
    if (bb == 0) {
	zset(c, 1);
	return;
    }
    if (bb > b)
	bb = b;
    zset(c, a);
    for (i=2; i<=bb; i++) {
	a--;
	zset(&u, a);
	zzmul(c, &u, &cc);
	zset(&u, i);
	zzdiv(&cc, &u, c, &v);
    }
}

/*
 * copy expression e1 to e2, collapsing along the way
 * on input, e2[0].d must point to valid index storage space
 * lastvalid is the last valid storage space in e2
 * on error (attempt to write past lastvalid), returns 1
 * on exit, eend points to last term in expression
 */
int ecopy(expr *e1, expr *e2, expr *lastvalid, expr **eend)
{
    int i, n;

    e2[0].isize = 0;
    if (e2 == lastvalid) {
	e2->next = NULL;
	*eend = e2;
	return 1;
    }
    e1 = e1->next;
    n = 0;
    while (e1 != NULL) {
	e2++;
	if (e2 == lastvalid) {
	    e2->next = NULL;
	    *eend = e2;
	    return 1;
	}
	e2[-1].next = e2;
	e2->d = e2[-1].d + n;
	n = e1->isize;
	e2->isize = n;
	for (i=0; i<n; i++)
	    e2->d[i] = e1->d[i];
	qcopy(&e1->coef, &e2->coef);
	e1 = e1->next;
    }
    e2->next = NULL;
    *eend = e2;
    return 0;
}

/* collapse away dead space from expression e, return ptr to new last term */
expr *collapsee(expr *e)
{
    int j, k;
    expr *last, *this;

    last = e;
    this = e[0].next;
    j = 0;
    while (this != NULL) {
	j++;
	if (this != e + j) {
	    e[j-1].next = e + j;
	    e[j] = *this;
	    e[j].d = e[j-1].d + e[j-1].isize;
	    for (k=0; k<this->isize; k++)
		e[j].d[k] = this->d[k];
	}
	last = this;
	this = this->next;
    }
    return e + j;
}

/*
 * add expression (term) f to e, returning -1 if error (exceeding maxx)
 * or 0 if okay
 * terms should be massaged (e.g. arranging indices in some canonical
 * order) before calling this routine
 */
int addtoe2(expr *f, expr *e, int maxx)
{
    expr ff;
    int i, j, k, m;
    qnumber z;
    expr *last, *this;

    if (qiszero(&(f->coef)))
	return 0;
    ff = *f;
/* NOTE with expr you should call some sort of expression-massager
before adding (e.g. with the gamma ijk, first sort ijk then sort each
triple so Gijk Gi'j'k' get swapped sometimes */
    ff.next = NULL;
    last = e;
    this = e->next;
    while (this != NULL) {
	if (ff.isize == this->isize) {
	    m = 1;
	    for (j=0; m && j<ff.isize; j++)
		if (ff.d[j] != this->d[j])
		    m = 0;
	    if (m) {
		qadd(&this->coef, &ff.coef, &z);
		if (qiszero(&z)) {
		    last->next = this->next;
		} else {
		    qcopy(&z, &this->coef);
		}
		return 0;
	    }
	}
	last = this;
	this = this->next;
    }
    if (last == e + maxx - 1) { /* expression is full, can't add new term */
				/* first try to collapse away dead space */
	if ((last = collapsee(e)) == e + maxx - 1)
	    return -1;
    }
    last->next = &last[1];
    last[1] = ff;
    last[1].d = last[0].d + last[0].isize;
    for (k=0; k<ff.isize; k++)
	last[1].d[k] = ff.d[k];
    return 0;
}

int fprterm(FILE *f, int s, int *d, int t)
{
    int i, ii, l;

    l = 0;
    switch(t) {
    case 1:
	for (i=0; i<s; i+=3) {
	    fprintf(f, " G_%s,%s,%s", V.cohname[d[i]],
			    V.cohname[d[i+1]], V.cohname[d[i+2]]);
	    l += 5 + V.namelen[d[i]] + V.namelen[d[i+1]] + V.namelen[d[i+3]];
	}
	break;
    case 2:
    case 3:
    case 4:
	for (ii=0; ii<s; ii+=ni) {
	    fprintf(f, " N(");
	    l += 3;
	    for (i=0; i<rkP; i++) {
		fprintf(f, "%d%c", d[ii+i], i==rkP-1 ? ';' : ',');
		l += NUMLEN(d[ii+i]) + 1;
	    }
	    for (i=rkP; i<ni; i++) {
		fprintf(f, "%d%c", d[ii+i], i==ni-1 ? ')' : ',');
		l += NUMLEN(d[ii+i]) + 1;
	    }
	}
	break;
    case 5:
	break;
    }
    return l;
}

int prterm(int s, int *d, int t)
{
    return fprterm(stdout, s, d, t);
}

/*
 * t = 1    equation of gammas (partial differential equation)
 * t = 2    equation of N's
 * t = 3    expression of N's
 * t = 4    expression of N's  with limit on length of print-out
 * t = 5    coefs only
 */
void prexpr(expr *e, int t)
{
    int i, j, l;
    char *s;
    double d, atof();
    qnumber z;
    expr *this;

    this = e->next;
    i = 1;
    l = 0;
    while (this != NULL) {
	if (t == 4 && l > 400) {
	    printf(" + ...\n");
	    return;
	}
	qcopy(&this->coef, &z);
	if (qsgn(&z) == 1) {
	    if (i == 0)
		printf(" + ");
	    else
		printf("  ");
	} else {
	    if (i == 0) {
		printf(" - ");
		qchs(&z);
	    } else
		printf(" ");
	}
	s = qtodec(&z);
	fputs(s, stdout);
	printf(" ");
	l += strlen(s) + 1;
	l += prterm(this->isize, this->d, t);
	this = this->next;
	i = 0;
    }
    if (t == 3 || t == 4 || t == 5)
	printf("\n");
    else
	printf(" = 0\n");
}

void reducecoef(expr *e)
{
    int dcleared;
    expr *this, *h;
    qnumber z, y;
    znumber a, b, g;

    if (e[0].next == NULL)
	return;
    dcleared = 0;
    while (!dcleared) {
	dcleared = 1;
	this = e[0].next;
	while (this != NULL && dcleared) {
	    if (this->coef.d.sign != 0 || this->coef.d.len != 1 ||
					    this->coef.d.s[0] != 1) {
		z.n = this->coef.d;
		z.d.sign = 0;
		z.d.len = 1;
		z.d.s[0] = 1;
		h = e[0].next;
		while (h != NULL) {
		    qqmul(&h->coef, &z, &y);
		    qcopy(&y, &h->coef);
		    h = h->next;
		}
		dcleared = 0;
	    } else {
		this = this->next;
	    }
	}
    }
/* now find gcd and divide out */
    zcopy(&e[0].next->coef.n, &g);
    this = e[0].next->next;
    while (this != NULL) {
	mgcd(&g, &this->coef.n, &a);
	zcopy(&a, &g);
	this = this->next;
    }
    this = e[0].next;
    while (this != NULL) {
	zzdiv(&this->coef.n, &g, &a, &b);
	zcopy(&a, &this->coef.n);
	this = this->next;
    }
}

#if 0
int checkclassassoc(variety *v, int r)
{
    int i, j, k, l, rk, p, q;
    qnumber x, y, z;

    rk = v->rkHstar;
    i = v->Ri[r];
    j = v->Rj[r];
    k = v->Rk[r];
    l = v->Rl[r];
    qzero(&z);
    for (p = 0; p < rk; p++) {
	for (q = 0; q < rk; q++) {
	    if (qsgn(&v->imatinvq[p*rk+q])) {
		qqmul(&v->potq[i*rk*rk+j*rk+p], &v->imatinvq[p*rk+q], &y);
		qqmul(&y, &v->potq[q*rk*rk+k*rk+l], &x);
		qadd(&z, &x, &y);
		qcopy(&y, &z);
		qqmul(&v->potq[j*rk*rk+k*rk+p], &v->imatinvq[p*rk+q], &y);
		qqmul(&y, &v->potq[q*rk*rk+i*rk+l], &x);
		qsub(&z, &x, &y);
		qcopy(&y, &z);
	    }
	}
    }
    return qsgn(&z) ? 1 : 0;
}
#endif

int checkclassassoc(variety *v, int r1, int r2, int otype)
{
    int i, j, k, l, rk, p, q, r, d, potd, xx, rv;
    int *pd, *g;
    qnumber denomq, w;
    znumber one, x, y, z, denom;

    rv = 0;
    rk = v->rkHstar;
    g = (int *)malloc(rk*rk*sizeof(int));
    pd = (int *)malloc(rk*rk*rk*sizeof(int));
    zset(&one, 1);
    zset(&denom, 1);
    for (i=0; i<rk*rk*rk; i++) {
	if (qsgn(&v->potq[i])) {
	    mgcd(&denom, &v->potq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->potq[i].d, &y, &denom);
	}
    }
    ztoq(&denom, &denomq);
    potd = ztoi(&denom);
    for (i=0; i<rk; i++) {
	for (j=0; j<rk; j++) {
	    for (k=0; k<rk; k++) {
		if (qsgn(&v->potq[i*rk*rk+j*rk+k])) {
		    qqmul(&v->potq[i*rk*rk+j*rk+k], &denomq, &w);
		    pd[i*rk*rk+j*rk+k] = ztoi(&w.n);
		} else {
		    pd[i*rk*rk+j*rk+k] = 0;
		}
	    }
	}
    }

    zset(&denom, 1);
    for (i=0; i<rk*rk; i++) {
	if (qsgn(&v->imatinvq[i])) {
	    mgcd(&denom, &v->imatinvq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->imatinvq[i].d, &y, &denom);
	}
    }
    ztoq(&denom, &denomq);
    d = ztoi(&denom);
    for (i=0; i<rk; i++) {
	for (j=0; j<rk; j++) {
	    qqmul(&v->imatinvq[i*rk+j], &denomq, &w);
	    g[i*rk+j] = ztoi(&w.n);
	}
    }

    for (r=r1; r<=r2; r++) {
	i = v->Ri[r];
	j = v->Rj[r];
	k = v->Rk[r];
	l = v->Rl[r];
	xx = 0;
	for (p = 0; p < rk; p++) {
	    for (q = 0; q < rk; q++) {
		if (g[p*rk+q]) {
		    xx += (pd[i*rk*rk+j*rk+p]*pd[q*rk*rk+k*rk+l] -
			    pd[j*rk*rk+k*rk+p]*pd[q*rk*rk+i*rk+l])*g[p*rk+q];
		}
	    }
	}
	if (xx)
	    rv = 1;
	if (xx || otype) {
	    printf("(%d)  (%d,%d,%d,%d)  / %d \\    associativity %s\n",
		    r+1, i+1, j+1, k+1, l+1, v->weight[r],
		    xx ? "fails" : "checks");
	}
    }
    free(g);
    free(pd);
    return rv;
}

/*
 * build specific relations, the indices of which are in rellist
 * (nlist of them)
 * on input e0 should be an empty expression
 */
int buildtheserels(variety *v, expr **rels, expr *e0, int *rellist,
int nlist, int maxexpr)
{
    int i, j, k, l, n, a, b, d, p, rk, where, xx, potd, ind[6];
    int *pd, *g;
    expr t;
    expr *e, *eend;
    znumber one, x, y, z, denom;
    qnumber denomq, w;

    t.d = ind;
    rk = v->rkHstar;
    g = (int *)malloc(rk*rk*sizeof(int));
    pd = (int *)malloc(rk*rk*rk*sizeof(int));



    zset(&one, 1);
    zset(&denom, 1);
    for (i=0; i<rk*rk*rk; i++) {
	if (qsgn(&v->potq[i])) {
	    mgcd(&denom, &v->potq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->potq[i].d, &y, &denom);
	}
    }
    ztoq(&denom, &denomq);
    potd = ztoi(&denom);
    for (i=0; i<rk; i++) {
	for (j=0; j<rk; j++) {
	    for (k=0; k<rk; k++) {
		if (qsgn(&v->potq[i*rk*rk+j*rk+k])) {
		    qqmul(&v->potq[i*rk*rk+j*rk+k], &denomq, &w);
		    pd[i*rk*rk+j*rk+k] = ztoi(&w.n);
		} else {
		    pd[i*rk*rk+j*rk+k] = 0;
		}
	    }
	}
    }

    zset(&denom, 1);
    for (i=0; i<rk*rk; i++) {
	if (qsgn(&v->imatinvq[i])) {
	    mgcd(&denom, &v->imatinvq[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v->imatinvq[i].d, &y, &denom);
	}
    }
    ztoq(&denom, &denomq);
    d = ztoi(&denom);

    for (i=0; i<rk; i++) {
	for (j=0; j<rk; j++) {
	    qqmul(&v->imatinvq[i*rk+j], &denomq, &w);
	    g[i*rk+j] = ztoi(&w.n);
	}
    }
    e = e0;
    where = 0;
    for (n=0; n<nlist; n++) {
	i = v->Ri[rellist[n]];
	j = v->Rj[rellist[n]];
	k = v->Rk[rellist[n]];
	l = v->Rl[rellist[n]];

	e->next = NULL;
	for (a=1; a<rk; a++) {
	    xx = 0;
	    for (p=0; p<rk; p++)
		xx += g[a*rk+p]*pd[i*rk*rk+j*rk+p];
	    qset(&t.coef, xx, d*potd);
	    t.isize = 3;
	    ind[0] = k;
	    ind[1] = l;
	    ind[2] = a;
	    massagedterm(&t);
	    if (addtoe2(&t, e, maxexpr-where)) {
		free(g);
		free(pd);
		return -1;
	    }
	    xx = 0;
	    for (p=0; p<rk; p++)
		xx += g[a*rk+p]*pd[j*rk*rk+k*rk+p];
	    qset(&t.coef, -xx, d*potd);
	    t.isize = 3;
	    ind[0] = i;
	    ind[1] = l;
	    ind[2] = a;
	    massagedterm(&t);
	    if (addtoe2(&t, e, maxexpr-where)) {
		free(g);
		free(pd);
		return -1;
	    }
	}
	for (a=1; a<rk; a++) {
	    xx = 0;
	    for (p=0; p<rk; p++)
		xx += g[a*rk+p]*pd[k*rk*rk+l*rk+p];
	    qset(&t.coef, xx, d*potd);
	    t.isize = 3;
	    ind[0] = i;
	    ind[1] = j;
	    ind[2] = a;
	    massagedterm(&t);
	    if (addtoe2(&t, e, maxexpr-where)) {
		free(g);
		free(pd);
		return -1;
	    }
	    xx = 0;
	    for (p=0; p<rk; p++)
		xx += g[a*rk+p]*pd[i*rk*rk+l*rk+p];
	    qset(&t.coef, -xx, d*potd);
	    t.isize = 3;
	    ind[0] = j;
	    ind[1] = k;
	    ind[2] = a;
	    massagedterm(&t);
	    if (addtoe2(&t, e, maxexpr-where)) {
		free(g);
		free(pd);
		return -1;
	    }
	}
	for (a=1; a<rk; a++) {
	    for (b=1; b<rk; b++) {
		qset(&t.coef, g[a*rk+b], d);
		t.isize = 6;
		ind[0] = i;
		ind[1] = j;
		ind[2] = a;
		ind[3] = k;
		ind[4] = l;
		ind[5] = b;
		massagedterm(&t);
		if (addtoe2(&t, e, maxexpr-where)) {
		    free(g);
		    free(pd);
		    return -1;
		}
		qset(&t.coef, g[a*rk+b], d);
		t.isize = 6;
		ind[0] = j;
		ind[1] = k;
		ind[2] = a;
		ind[3] = i;
		ind[4] = l;
		ind[5] = b;
		qchs(&t.coef);
		massagedterm(&t);
		if (addtoe2(&t, e, maxexpr-where)) {
		    free(g);
		    free(pd);
		    return -1;
		}
	    }
	}
	eend = collapsee(e);
	reducecoef(e);
	rels[n] = e;
	if (n != nlist-1) {
	    if (eend == e0 + maxexpr - 1) {
		free(g);
		free(pd);
		return -1;
	    }
	    where = eend + 1 - e0;
	    eend[1].d = eend[0].d + eend[0].isize;
	    eend[1].next = NULL;
	    e = eend + 1;
	}
    }
    free(g);
    free(pd);
    return 0;
}

/*
 * build differential realtions for variety v
 * if (i,j,k,l) yields a useful relation then its weight is
 *   codim(y_i) + codim(y_j) + codim(y_k) + codim(y_l) - dim
 */
int buildrels(variety *v)
{
    int a, i, j, k, l, rk, n;

    rk = v->rkHstar;
    a = rk - 1;
    v->nrels = a*(a-1)*(a*a-a+2)/8;
    v->Ri = (int *)malloc(v->nrels*sizeof(int));
    v->Rj = (int *)malloc(v->nrels*sizeof(int));
    v->Rk = (int *)malloc(v->nrels*sizeof(int));
    v->Rl = (int *)malloc(v->nrels*sizeof(int));
    v->weight = (int *)malloc(v->nrels*sizeof(int));
    n = 0;
    for (i=1; i<rk; i++) {
	for (j=i+1; j<rk; j++) {
	    for (k=j+1; k<rk; k++) {
		for (l=k+1; l<rk; l++) {
		    v->Ri[n] = i;
		    v->Rj[n] = j;
		    v->Rk[n] = k;
		    v->Rl[n] = l;
		    v->weight[n] = v->cohcodim[i] + v->cohcodim[j] +
				v->cohcodim[k] + v->cohcodim[l] - v->dim;
		    n++;
		    v->Ri[n] = i;
		    v->Rj[n] = j;
		    v->Rk[n] = l;
		    v->Rl[n] = k;
		    v->weight[n] = v->cohcodim[i] + v->cohcodim[j] +
				v->cohcodim[k] + v->cohcodim[l] - v->dim;
		    n++;
		    v->Ri[n] = i;
		    v->Rj[n] = k;
		    v->Rk[n] = j;
		    v->Rl[n] = l;
		    v->weight[n] = v->cohcodim[i] + v->cohcodim[j] +
				v->cohcodim[k] + v->cohcodim[l] - v->dim;
		    n++;
		}
	    }
	}
    }
    for (i=1; i<rk; i++) {
	for (k=i+1; k<rk; k++) {
	    for (l=k+1; l<rk; l++) {
		v->Ri[n] = i;
		v->Rj[n] = i;
		v->Rk[n] = k;
		v->Rl[n] = l;
		v->weight[n] = v->cohcodim[i] + v->cohcodim[i] +
			    v->cohcodim[k] + v->cohcodim[l] - v->dim;
		n++;
		v->Ri[n] = i;
		v->Rj[n] = k;
		v->Rk[n] = k;
		v->Rl[n] = l;
		v->weight[n] = v->cohcodim[i] + v->cohcodim[k] +
			    v->cohcodim[k] + v->cohcodim[l] - v->dim;
		n++;
		v->Ri[n] = i;
		v->Rj[n] = k;
		v->Rk[n] = l;
		v->Rl[n] = l;
		v->weight[n] = v->cohcodim[i] + v->cohcodim[k] +
			    v->cohcodim[l] + v->cohcodim[l] - v->dim;
		n++;
	    }
	}
    }
    for (i=1; i<rk; i++) {
	j = i;
	for (k=i+1; k<rk; k++) {
	    l = k;
	    v->Ri[n] = i;
	    v->Rj[n] = j;
	    v->Rk[n] = k;
	    v->Rl[n] = l;
	    v->weight[n] = v->cohcodim[i] + v->cohcodim[j] +
			v->cohcodim[k] + v->cohcodim[l] - v->dim;
	    n++;
	}
    }
    if (n != v->nrels)
	printf("error\n");
}

int dowehave(int *d, qnumber *v)
{
    int i, j, k;

    j = 0;
    for (i=0; i<nknown; i++) {
	j = 1;
	for (k=0; k<ni && j; k++)
	    if (d[k] != vallist[i*ni+k])
		j = 0;
	if (j)
	    break;
    }
    if (j) {
	qcopy(&value[i], v);
	return 1;
    } else {
	return 0;
    }
}

int havelin(int *d, expr **v)
{
    int i, j, k;

    j = 0;
    for (i=0; i<nlinear; i++) {
	j = 1;
	for (k=0; k<ni && j; k++)
	    if (d[k] != linlist[i*ni+k])
		j = 0;
	if (j)
	    break;
    }
    if (j) {
	*v = linrhs[i];
	return 1;
    } else {
	return 0;
    }
}

void removeknowledge(int *d)
{
    int i, j, k, l;

    j = 0;
    for (i=0; i<nknown; i++) {
	j = 1;
	for (k=0; k<ni && j; k++)
	    if (d[k] != vallist[i*ni+k])
		j = 0;
	if (j)
	    break;
    }
    if (j) {
	for (l=i; l<nknown-1; l++) {
	    for (k=0; k<ni; k++)
		vallist[l*ni+k] = vallist[(l+1)*ni+k];
	    qcopy(&value[l+1], &value[l]);
	}
	nknown--;
    }
}

void addknowledge(int *d, qnumber *v)
{
    int i;

    if (nknown == MAXKNOWN) {
	printf("list full\n");      /* DO SOMETHING MORE ELEGANT? */
	return;
    }
    for (i=0; i<ni; i++)
	vallist[nknown*ni+i] = d[i];
    qcopy(v, &value[nknown]);
    nknown++;
}

int islinear(expr *e, int *len)
{
    int l;
    expr *this;

    l = 0;
    for (this = e[0].next; this != NULL; this = this->next) {
	if (this->isize == 2*ni)
	    return 0;
	l++;
    }
    *len = l;
    return 1;
}

/*
 * for variety v and multiindex        c_1, .. c_rkPic
 *                                d_rkH2+2, .. d_rkHstar
 * satisfying      sum c_j A_j  -  sum d_j (codim(y_j)-1)  =  weight(reln)
 * where A_j = int_(B_j)^ c_1(TV)  (the "c1t" numbers)
 * compute the equation which is the coefficient of
 *   exp(.. y1) .. exp(.. yr) prod y_j^d_j
 * in the given relation; the relation should consist of terms
 * G_a1,a2,a3 with codim(y_a1)+codim(y_a2)+codim(y_a3)-dimV = weight
 * and G_a1,a2,a3 G_b1,b2,b3 with
 * codim(y_a1)+codim(y_a2)+codim(y_a3)-dimV + codim(y_a1)+codim(y_a2)+
 *                                              codim(y_a3)-dimV = weight
 * mode = 0 for normal operation, mode = 1 to set N(valind)=newval
 * (if la=1, this is the way to force a new N value since it updates
 * the linear relationships)
 * returns
 *  -1: error, had to trucate terms (try increasing MAXEXPR)
 *  -2: error, overflow occurred
 *  -3: error, interrupt
 *  -4: error, overflow occurred in the critical section of linear algebra
 *      such that linear relations in memory may be garbled
 *   0: relation is trivial
 *   1: relation doesn't yield any check
 *   2: relation checks
 *   3: relation doesn't check
 *   5: relation determines an integer value for N(c0,d0)
 *      this value has been placed in the N array
 *   8: did linear algebra, but relation to add exceeded term limit
 *   9: did linear algebra, but ran out of memory doing manipulations,
 *      so relation had no effect
 *  10: did linear algebra, got nnew new #s and new linear relationship
 *  11: did linear algebra, got nnew new #s
 *  12: did linear algebra, got new linear relationship
 */
int doreln(variety *v, expr *r, int *c, int *d, int *valind, qnumber *newval,
int la, int termlim, int verbose, int mode, int *nnew)
{
    int i, j, k, il, m, t, a1, a2, a3, b1, b2, b3, rk, rkH2, tt, any, rv;
    int where, nadded, wta, wtb, da, db, multii, iidiv, oov, oldnknown;
    int *indspace, *scr, *dspace, *ccl, *mu, *mua, *mub;
    int *iia, *iib, *ii0, *iidot;
    expr e[MAXEXPR], f;
    expr *first, *last, *this, *lthis;
    expr *vterm, *cterm, *linexpr, *e1, *e2, *eend, *junk;
    qnumber y, yy, z, za, zb, zn, newv;
    znumber z1, z2, z3;

    oov = ov;
    ov = 0;
    tt = 0;
    rk = v->rkHstar;
    rkH2 = v->rkH2;
    multii = rkP + rk - rkH2 - 2;
    indspace = (int *)malloc(MAXEXPR*2*ni*sizeof(int));
    scr = (int *)malloc((2*ni + rkP*rkH2 + 3*rk + 4*multii)*sizeof(int));
    dspace = scr;
    ccl = dspace + 2*ni;
    mu = ccl + rkP*rkH2;
    mua = mu + rk;
    mub = mua + rk;
    ii0 = mub + rk;
    iia = ii0 + multii;
    iib = iia + multii;
    iidot = iib + multii;
    e[0].isize = 0;
    e[0].d = indspace;
    e[0].next = NULL;
    f.d = dspace;
    if (mode == 0) {
	for (i=0; i<rkP; i++)
	    for (j=0; j<rkH2; j++)
		ccl[i*rkH2+j] = v->cclass[i*rkH2+j];
	mu[0] = 0;
	for (j=0; j<rkH2; j++)
	    mu[j+1] = 0;
	for (j=0; j<rkH2; j++)
	    for (i=0; i<rkP; i++)
		mu[j+1] += c[i]*ccl[i*rkH2+j];
	for (j=rkH2+1; j<rk; j++)
	    mu[j] = v->cccd;
	for (i=0; i<rkP; i++)
	    ii0[i] = c[i];
	for (i=1; i<ni-rkP; i++)
	    ii0[i-1+rkP] = d[i];
	for (i=0; i<rkP; i++)
	    iidot[i] = v->c1t[i];
	iidiv = v->cohcodim[rkH2 + 1] - 1;
	for (i=1; i<ni-rkP; i++)
	    iidot[i-1+rkP] = -(v->cohcodim[rkH2 + 1 + i] - 1);
	for (this = r->next; this != NULL; this = this->next) {
	    if (nsig) {
		free(indspace);
		free(scr);
		ov = (ov || oov);
		return -3;
	    }
	    if (this->isize == 0) {
printf("shouldn't get here -- classical associativity disallows this!\n");
exit(1);
	    } else if (this->isize == 3) {
		a1 = this->d[0];
		a2 = this->d[1];
		a3 = this->d[2];
		qset(&z, mu[a1]*mu[a2]*mu[a3], v->cccd*v->cccd*v->cccd);
		qqmul(&z, &this->coef, &f.coef);
		f.isize = ni;
		for (i=0; i<rkP; i++)
		    f.d[i] = c[i];
		for (i=0; i<ni-rkP; i++)
		    f.d[rkP+i] = d[i];
		if (a1 >= rkH2+1)
		    f.d[a1-(rkH2+1)+rkP]++;
		if (a2 >= rkH2+1)
		    f.d[a2-(rkH2+1)+rkP]++;
		if (a3 >= rkH2+1)
		    f.d[a3-(rkH2+1)+rkP]++;
		massagenterm(&f);
/* it's actually FASTER to collect all terms and then substitute numbers */
/* than to substitute numbers as we go along in generating the relation */
#if 0
		if (!qiszero(&f.coef)) {
		    if (!verbose && dowehave(f.d, &zn)) {
			qqmul(&f.coef, &zn, &zb);
			qcopy(&zb, &f.coef);
			f.isize = 0;
		    }
		    if (addtoe2(&f, e, MAXEXPR)) {
			tt = 1;
		    }
		}
#else
		if (addtoe2(&f, e, MAXEXPR)) {
		    tt = 1;
		}
#endif
	    } else {
		a1 = this->d[0];
		a2 = this->d[1];
		a3 = this->d[2];
		b1 = this->d[3];
		b2 = this->d[4];
		b3 = this->d[5];
		wta = v->cohcodim[a1] + v->cohcodim[a2] + v->cohcodim[a3] -
								v->dim;
		wtb = v->cohcodim[b1] + v->cohcodim[b2] + v->cohcodim[b3] -
								v->dim;
/* printf("%d %d\n", wta, wtb); */
/* exhaust over 0 <= (all indices but first "d") <= (c,d) */
		for (i=0; i<multii; i++)
		    iia[i] = 0;
		for (i=0; i<rkP; i++)
		    iia[i] = v->loc[i];
		for (i=0; i<multii; i++)
		    iib[i] = ii0[i] - iia[i];
		while (1) {
		    if (nsig) {
			free(indspace);
			free(scr);
			ov = (ov || oov);
			return -3;
		    }
/* for (i=0; i<multii; i++) printf("%d ", iia[i]); printf("   ");
for (i=0; i<multii; i++) printf("%d ", iib[i]); printf("   ");
for (i=0; i<multii; i++) printf("%d ", ii0[i]); printf("\n"); */
/* compute d's, if integers in range, do (iia,iib) */
		    j = -wta;
		    for (i=0; i<multii; i++)
			j += iia[i]*iidot[i];
		    if (j >= 0 && j%iidiv == 0 && (da=j/iidiv) <= d[0]) {
/*  printf("  YES!!\n");  */
			db = d[0] - da;
			mua[0] = 0;
			for (j=0; j<rkH2; j++)
			    mua[j+1] = 0;
			for (j=0; j<rkH2; j++)
			    for (i=0; i<rkP; i++)
				mua[j+1] += iia[i]*ccl[i*rkH2+j];
			for (j=rkH2+1; j<rk; j++)
			    mua[j] = v->cccd;
			mub[0] = 0;
			for (j=0; j<rkH2; j++)
			    mub[j+1] = mu[j+1] - mua[j+1];
			for (j=rkH2+1; j<rk; j++)
			    mub[j] = v->cccd;
			qset(&za, mua[a1]*mua[a2]*mua[a3],
					v->cccd*v->cccd*v->cccd);
			qset(&zb, mub[b1]*mub[b2]*mub[b3],
					v->cccd*v->cccd*v->cccd);
			qqmul(&za, &zb, &z);
			cmb(d[0], da, &z1);
			for (i=rkP; i<ni-1; i++) {
			    cmb(ii0[i], iia[i], &z2);
			    zzmul(&z1, &z2, &z3);
			    zcopy(&z3, &z1);
			}
			ztoq(&z1, &y);
			qqmul(&z, &y, &yy);
			qqmul(&yy, &this->coef, &f.coef);
			for (i=0; i<rkP; i++)
			    f.d[i] = iia[i];
			f.d[rkP] = da;
			for (i=rkP; i<ni-1; i++)
			    f.d[i+1] = iia[i];
			for (i=0; i<rkP; i++)
			    f.d[ni+i] = iib[i];
			f.d[ni+rkP] = db;
			for (i=rkP; i<ni-1; i++)
			    f.d[ni+i+1] = iib[i];
			if (a1 >= rkH2+1)
			    f.d[a1-(rkH2+1)+rkP]++;
			if (a2 >= rkH2+1)
			    f.d[a2-(rkH2+1)+rkP]++;
			if (a3 >= rkH2+1)
			    f.d[a3-(rkH2+1)+rkP]++;
			if (b1 >= rkH2+1)
			    f.d[ni+b1-(rkH2+1)+rkP]++;
			if (b2 >= rkH2+1)
			    f.d[ni+b2-(rkH2+1)+rkP]++;
			if (b3 >= rkH2+1)
			    f.d[ni+b3-(rkH2+1)+rkP]++;
			f.isize = 2*ni;
/* printf("adding %s\n", qtodec(&f.coef)); */
			massagenterm(&f);
/* it's actually FASTER to collect all terms and then substitute numbers */
/* than to substitute numbers as we go along in generating the relation */
#if 0
			if (!qiszero(&f.coef)) {
			    if (0&&/* TEST */ !verbose) {
				if (dowehave(f.d, &zn)) {
				    if (dowehave(f.d+ni, &zb)) {
					qqmul(&f.coef, &zn, &za);
					qqmul(&za, &zb, &f.coef);
					f.isize = 0;
				    } else {
					qqmul(&f.coef, &zn, &zb);
					qcopy(&zb, &f.coef);
					for (i=0; i<ni; i++)
					    f.d[i] = f.d[ni+i];
					f.isize = ni;
				    }
				} else if (dowehave(f.d+ni, &zn)) {
				    qqmul(&f.coef, &zn, &zb);
				    qcopy(&zb, &f.coef);
				    f.isize = ni;
				}
			    }
			    if (addtoe2(&f, e, MAXEXPR)) {
				tt = 1;
			    }
			}
#else
			if (addtoe2(&f, e, MAXEXPR)) {
			    tt = 1;
			}
#endif
		    }   /*  else printf("   NO!!\n");  */
/* increment to next */
/*
printf("B ");
for (i=0; i<multii; i++) printf("%d ", iia[i]); printf("   ");
for (i=0; i<multii; i++) printf("%d ", iib[i]); printf("   ");
for (i=0; i<multii; i++) printf("%d ", ii0[i]); printf("\n");
*/
		    for (i=multii-1; i>=0 && iib[i] == ((i<rkP)?v->loc[i]:0);
								    i--)
			;
		    if (i < 0)          /* we're done */
			break;
		    iib[i]--;
		    iia[i]++;
		    for (j=i+1; j<multii; j++) {
			iia[j] = ((j<rkP)?v->loc[j]:0);
			iib[j] = ii0[j]-iia[j];
		    }
/*
printf("A ");for (i=0; i<multii; i++) printf("%d ", iia[i]); printf("   ");
for (i=0; i<multii; i++) printf("%d ", iib[i]); printf("   ");
for (i=0; i<multii; i++) printf("%d ", ii0[i]); printf("\n");
*/
		}
	    }
	}
	if (verbose) {
	    printf("Relation   ");
	    for (i=0; i<rkP; i++)
		printf(" %d", c[i]);
	    printf(" ;");
	    for (i=0; i<ni-rkP; i++)
		printf(" %d", d[i]);
	}
	if (e[0].next == NULL) {
	    if (verbose)
		printf("  is trivial\n");
	    if (ov) {
		if (verbose)
		    printf("warning: overflow occured at some point\n");
		free(scr);
		free(indspace);
		ov = (ov || oov);
		return -2;
	    }
	    if (tt) {
		free(scr);
		free(indspace);
		ov = oov;
		return -1;
	    }
	    free(scr);
	    free(indspace);
	    ov = oov;
	    return 0;
	}
	if (verbose) {
	    printf(" :\n");
	    prexpr(e,2);
	}




/* now substitute */
	any = 1;
	while (any) {
	    any = 0;
	    last = e;
	    this = e[0].next;
	    while (this != NULL && !any) {
		if (nsig) {
		    free(indspace);
		    free(scr);
		    ov = (ov || oov);
		    return -3;
		}
		if (this->isize == ni) {
		    if (dowehave(this->d, &zn)) {
			qqmul(&this->coef, &zn, &f.coef);
			f.isize = 0;
			last->next = this->next;
			if (addtoe2(&f, e, MAXEXPR) == -1) {
/* shouldn't get here since numerical substitutions shorten expressions */
			    printf("overloaded 1\n");
			    exit(1);
			}
			any = 1;
		    }
		} else if (this->isize == 2*ni) {
		    if (dowehave(this->d, &zn)) {
			qqmul(&this->coef, &zn, &f.coef);
			f.isize = ni;
			for (i=0; i<ni; i++)
			    f.d[i] = this->d[ni+i];
			last->next = this->next;
			if (addtoe2(&f, e, MAXEXPR) == -1) {
/* shouldn't get here since numerical substitutions shorten expressions */
			    printf("overloaded 2\n");
			    exit(1);
			}
			any = 1;
		    } else if (dowehave(this->d+ni, &zn)) {
			qqmul(&this->coef, &zn, &f.coef);
			f.isize = ni;
			for (i=0; i<ni; i++)
			    f.d[i] = this->d[i];
			last->next = this->next;
			if (addtoe2(&f, e, MAXEXPR) == -1) {
/* shouldn't get here since numerical substitutions shorten expressions */
			    printf("overloaded 3\n");
			    exit(1);
			}
			any = 1;
		    }
		}
		last = this;
		this = this->next;
	    }
	}
    } else {
	e[0].next = &e[1];
	e[1].d = e[0].d;
	e[1].isize = ni;
	for (i=0; i<ni; i++)
	    e[1].d[i] = valind[i];
	qset(&e[1].coef, 1, 1);
	e[1].next = &e[2];
	e[2].d = e[1].d + ni;
	e[2].isize = 0;
	e[2].next = NULL;
	qcopy(newval, &e[2].coef);
	qchs(&e[2].coef);
    }

    if (la && islinear(e, &i)) {    /* if linear, substitute linear terms */
	any = 1;
	while (any) {
	    any = 0;
	    last = e;
	    this = e[0].next;
	    while (this != NULL && !any && !tt) {
		if (nsig) {
		    free(indspace);
		    free(scr);
		    ov = (ov || oov);
		    return -3;
		}
		if (this->isize == ni) {
		    if (havelin(this->d, &linexpr)) {
			qcopy(&this->coef, &zn);
			last->next = this->next;
			for (lthis = linexpr[0].next; lthis != NULL;
						    lthis = lthis->next) {
			    qqmul(&zn, &lthis->coef, &f.coef);
			    f.isize = lthis->isize;
			    for (i=0; i<f.isize; i++)
				f.d[i] = lthis->d[i];
			    if (addtoe2(&f, e, MAXEXPR) == -1)
				tt = 1;
			}
			any = 1;
		    }
		}
		last = this;
		this = this->next;
	    }
	}
    }

    if (e[0].next == NULL) {
	if (verbose)
	    printf("Relation checks!\n");
	rv = 2;
    } else {
	if (verbose) {
	    if (la)
		printf(
"After substituting known numbers and relations, this simplifies to:\n");
	    else
		printf(
"After substituting known numbers, this simplifies to:\n");
	    prexpr(e,2);
	}
	if (la && !tt && !ov && (il=islinear(e, &i)) &&
			(mode == 1 || termlim == 0 || i <= termlim)) {
	    nadded = 0;
	    if ((e[0].next)->next == NULL && e[0].next->isize == 0) {
		if (verbose)
		    printf("Relation FAILS!\n");
		rv = 3;
	    } else {        /* isolate one term, backsub, and cleanup */
		first = e[0].next;
		this = first->next;
		if (first->isize == 0) {
		    qcopy(&first->coef, &za);
		    qcopy(&this->coef, &first->coef);
		    qcopy(&za, &this->coef);
		    first->isize = ni;
		    first->d = this->d;
		    this->isize = 0;
		    this->d += ni;
		}
		qcopy(&first->coef, &za);
		qchs(&za);
		while (this != NULL) {
		    qqdiv(&this->coef, &za, &zb);
		    qcopy(&zb, &this->coef);
		    this = this->next;
		}
/* relation is now N(first->d) = (rest of expression e) */

/* backsubstitute, copying from main linear expr buffer to alternate */
		linex2[0].isize = 0;
		linex2[0].d = linex2d;
		linex2[0].next = NULL;
		e2 = linex2;
		where = 0;
		for (i=0; i<nlinear; i++) {
		    if (nsig) {
			free(indspace);
			free(scr);
			ov = (ov || oov);
			return -3;
		    }
		    if (ecopy(linrhs[i], e2, linex2 + MAXLINEXPR - 1, &junk)) {
			break;
		    } else {
			last = e2;
			this = e2->next;
			while (this != NULL) {
			    if (this->isize == ni) {
				for (j=0; j<ni && this->d[j]==first->d[j]; j++)
				    ;
				if (j == ni) {
				    qcopy(&this->coef, &zn);
				    last->next = this->next;
				    for (lthis = first->next; lthis != NULL;
							lthis = lthis->next) {
					qqmul(&zn, &lthis->coef, &f.coef);
					f.isize = lthis->isize;
					for (k=0; k<f.isize; k++)
					    f.d[k] = lthis->d[k];
					if (addtoe2(&f, e2, MAXLINEXPR - where)
								    == -1) {
					    goto bailout;
					}
				    }
				}
			    }
			    last = this;
			    this = this->next;
			}
			eend = collapsee(e2);
			linrhs2[i] = e2;
			if (eend == linex2 + MAXLINEXPR - 1) {
			    break;
			}
			where = eend + 1 - linex2;
			eend[1].d = eend[0].d + eend[0].isize;
			eend[1].next = NULL;
			e2 = eend + 1;
		    }
		}
bailout:
/* add current relation */
		if (i < nlinear || nlinear == MAXLIN) {
		    rv = 9;
		} else {
		    for (i=0; i<ni; i++)
			linlist[nlinear*ni + i] = first->d[i];
		    first->isize = 0;
		    first->d += ni;
		    if (ecopy(first, e2, linex2 + MAXLINEXPR - 1, &junk)) {
			rv = 9;
		    } else if (!ov) {
			linrhs2[nlinear] = e2;
			nlinear++;
/* clean up and copy back */
			linex[0].isize = 0;
			linex[0].d = linexd;
			linex[0].next = NULL;
			e1 = linex;
			where = 0;
			i = 0;
			oldnknown = nknown;
			while (i<nlinear) {
			    this = linrhs2[i];
			    if (this->next == NULL) {
				if (ov == 0 && tt == 0) {
				    qzero(&za);
				    addknowledge(&linlist[i*ni], &za);
				    nadded++;
				}
				for (j=i; j<nlinear-1; j++) {
				    linrhs2[j] = linrhs2[j+1];
				    for (k=0; k<ni; k++)
					linlist[j*ni+k] = linlist[(j+1)*ni+k];
				}
				m = 0;
				nlinear--;
			    } else if ((this->next)->next == NULL &&
						(this->next)->isize == 0) {
				if (ov == 0 && tt == 0) {
				    addknowledge(&linlist[i*ni],
						    &(this->next)->coef);
				    nadded++;
				}
				for (j=i; j<nlinear-1; j++) {
				    linrhs2[j] = linrhs2[j+1];
				    for (k=0; k<ni; k++)
					linlist[j*ni+k] = linlist[(j+1)*ni+k];
				}
				m = 0;
				nlinear--;
			    } else {
				if (ecopy(this, e1, linex + MAXLINEXPR - 1,
								&eend)) {
/* shouldn't get here since we're shortening expressions, if anything */
j=0;
while (this != NULL) { j++; this = this->next; }
				    printf(
"overloaded 4  %d  %d  %d  %d  %d\n", i, nlinear, e1-linex, j, junk - linex2);
				    exit(1);
				}
				linrhs[i] = e1;
				if (i != nlinear-1) {
				    if (eend == linex + MAXLINEXPR - 1) {
/* shouldn't get here since we're shortening expressions, if anything */
					printf(
"overloaded 5  %d  %d\n", i, nlinear);
					exit(1);
				    }
				    eend[1].d = eend[0].d + eend[0].isize;
				    eend[1].next = NULL;
				    e1 = eend + 1;
				}
				m = 1;
				i++;
			    }
			}
			if (m) {
			    this = linrhs[nlinear - 1];
			    while (this->next != NULL)
				this = this->next;
			    if (nlinear == MAXLIN ||
				    this - linex >= MAXLINEXPR - LBUFFER - 1) {
				m = 0;
				nlinear--;
			    }
			}
			if (nadded > 0 && m)
			    rv = 10;
			else if (nadded > 0 && !m)
			    rv = 11;
			else if (nadded == 0 && m)
			    rv = 12;
			else
			    rv = 9;
			*nnew = nadded;
			if (ov) {
			    if (v) {
				printf(
"warning: overflow occured at critical point of linear algebra\n");
				printf(
"linear relations in memory are garbled\n");
				printf(
"clearing all knowledge of linear relations\n");
			    }
			    nknown = oldnknown;
			    nlinear = 0;
			    free(scr);
			    free(indspace);
			    return -4;
			}
		    }
		}
	    }
	} else {
	    if ((e[0].next)->next == NULL) {
		this = e[0].next;
		if (this->isize == 0) {
		    if (verbose)
			printf("Relation FAILS!\n");
		    rv = 3;
		} else if (this->isize == ni) {
		    if (verbose) {
			printf("This determines ");
			prterm(1, this->d, 2);
			printf(" = 0\n");
		    }
		    if (ov) {
			if (verbose)
			    printf("Because of overflow, can't trust this.\n");
			free(scr);
			free(indspace);
			return -2;
		    } else if (tt) {
			printf(
"Because terms were truncated, can't trust this.\n");
			free(scr);
			free(indspace);
			ov = oov;
			return -1;
		    } else {
			qzero(&newv);
			addknowledge(this->d, &newv);
			free(scr);
			free(indspace);
			*nnew = 1;
			ov = oov;
			return 5;
		    }
		} else {
		    rv = 1;
		}
	    } else if (((e[0].next)->next)->next == NULL) {
		last = e[0].next;
		this = last->next;
		j = 0;
		if (last->isize == 0 && this->isize == ni) {
		    cterm = last;
		    vterm = this;
		    j = 1;
		} else if (last->isize == ni && this->isize == 0) {
		    cterm = this;
		    vterm = last;
		    j = 1;
		}
		if (j) {
		    qqdiv(&cterm->coef, &vterm->coef, &newv);
		    qchs(&newv);
		    if (verbose) {
			printf("This determines ");
			prterm(1, vterm->d, 2);
			printf(" = %s\n", qtodec(&newv));
		    }
		    if (ov) {
			if (verbose)
			    printf("Because of overflow, can't trust this.\n");
			free(scr);
			free(indspace);
			return -2;
		    } else if (tt) {
			printf(
"Because terms were truncated, can't trust this.\n");
			free(scr);
			free(indspace);
			ov = oov;
			return -1;
		    } else {
			addknowledge(vterm->d, &newv);
#if 0
			for (i=0; i<ni; i++)
			    valind[i] = vterm->d[i];
#endif
			free(scr);
			free(indspace);
			*nnew = 1;
			ov = oov;
			return 5;
		    }
		} else {
		    rv = 1;
		}
	    } else {
		if (la && il)
		    rv = 8;
		else
		    rv = 1;
	    }
	}
    }
    if (ov) {
	if (v)
	    printf("warning: overflow occured at some point\n");
	free(scr);
	free(indspace);
	return -2;
    }
    if (tt) {
	free(scr);
	free(indspace);
	ov = oov;
	return -1;
    }





    free(scr);
    free(indspace);
    ov = oov;
    return rv;
}

int rfind(int a[4])
{
    int i, j, t;

    for (i=0; i<V.nrels; i++) {
	if (a[0] == V.Ri[i] && a[1] == V.Rj[i] && a[2] == V.Rk[i] &&
						    a[3] == V.Rl[i])
	    return i+1;
    }
    return 0;
}

/* return ss=0 for trivial, else ss=+1/-1 and relation in std form in b */
void rnorm(int a[4], int b[4], int *ss)
{
    int s, m, u, v;

    if (a[0] == a[2] || a[1] == a[3]) {
	b[0] = 0;
	b[1] = 0;
	b[2] = 0;
	b[3] = 0;
	*ss = 0;
	return;
    }

    /* first get smallest to first position */
    m = a[0];
    u = 0;
    for (v=1; v<4; v++) {
	if (a[v] < m) {
	    m = a[v];
	    u = v;
	}
    }
    if (u == 0) {
	b[0] = a[0];
	b[1] = a[1];
	b[2] = a[2];
	b[3] = a[3];
    } else if (u == 1) {
	b[0] = a[1];
	b[1] = a[0];
	b[2] = a[3];
	b[3] = a[2];
    } else if (u == 2) {
	b[0] = a[2];
	b[1] = a[3];
	b[2] = a[0];
	b[3] = a[1];
    } else {
	b[0] = a[3];
	b[1] = a[2];
	b[2] = a[1];
	b[3] = a[0];
    }

    /* swap if needed to get b[1] less than b[3] */
    if (b[1] > b[3]) {
	s = -1;
	u = b[1];
	b[1] = b[3];
	b[3] = u;
    } else {
	s = 1;
    }

    /* one case needs more fixing */
    if (b[0] == b[1] && b[2] > b[3]) {
	u = b[2];
	b[2] = b[3];
	b[3] = u;
    }

    *ss = s;
    return;
}



int mainloop(int first, int readfile, FILE *thefile)
{
    static int nrels, reli, verbose, iorder, autod, linalg, termlim;
    static int rno[MAXNRELS], relind[6*MAXRELEXPR], nr;
    static int *wt;
    static expr drel[MAXRELEXPR], *drlist[MAXNRELS];
    int ch, c, d, r, c0, c1, d0, d1, cc, dd, whichrel, nadded, otermlim;
    int i, j, jj, k, kk, l, ll, m, n, any, tt, cdota, junk[4], junk2[4];
    int *inp, *outp, *cc1, *cc2, *dd1, *dd2, *ccc, *ddd;
    time_t ti;
    struct tms tis;
    clock_t tisc;
    qnumber a, b, z;
    char li[1000], ss;
    expr *this2;
    char *u, *p, *p2, *q;
    FILE *outfile, *thenewfile;

    if (first) {
	verbose = 0;
	iorder = 0;
	linalg = 0;
	termlim = 0;
	linlist = NULL;
	autod = 0;
	readfile = 0;
	V.dim = 0;
    }
    nsig = 0;
    signal(SIGINT, userinputsig);
    if (readfile) {
	if (ov)
	    printf("Overflow!\n");
    } else {
	printf(ov?"OV!>":">");
    }
    while ((i=fgetline(readfile ? thefile:stdin, li, 999)) != EOF) {
	signal(SIGINT, mysig);
	u = li;
	while (*u != '\0' && (isspace(*u) || *u == '>'))
	    u++;
	if (*u != '\0') {
	    ch = toupper(*u);
	    if (*u == '*') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    if (V.dim == 2 && V.rkPic == 2) {
			for (j = 10; j >= V.loc[1]; j--) {
			    for (i = V.loc[0]; i<=15; i++) {
				k = i*V.c1t[0] + j*V.c1t[1] - 1;
				if (i == 0 && j == 0) {
				    putchar('*');
				} else if (k < 0) {
				    putchar('X');
				} else {
				    junk[0] = i;
				    junk[1] = j;
				    junk[2] = k;
				    l = dowehave(junk, &z);
				    if (l) {
					l = qsgn(&z);
					if (l == 0)
					    putchar('0');
					else if (l > 0)
					    putchar('+');
					else
					    putchar('-');
				    } else {
					putchar(' ');
				    }
				}
			    }
			    putchar('\n');
			}
		    } else {
			printf(
"implemented only for dimension 2, rank Picard 2\n");
		    }
		}
	    } else if (*u == '<') {
		if (readfile) {
		    printf(
"Break to user; you may view data or make changes\n");
		    printf(
"Enter \"W\" to resume from file, or \"Z\" to close file and proceed.\n");
		    if (mainloop(0, 0, stdin) == 2)
			return 0;
		} else {
		    printf("Command \">\" only valid from a file\n");
		}
	    } else if (*u == '#') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    inp = (int *)malloc(V.rkPic*sizeof(int));
		    p = &u[1];
		    j = 1;
		    for (i=0; i<V.rkPic; i++) {
			while (*p != '\0' && isspace(*p))
			    p++;
			if (*p == '\0') {
			    printf("not enough input\n");
			    j = 0;
			    break;
			} else {
			    q = p;
			    while (*q != '\0' && !isspace(*q))
				q++;
			    if (*q == '\0') {
				inp[i] = atoi(p);
				p = q;
			    } else {
				*q = '\0';
				inp[i] = atoi(p);
				p = q + 1;
			    }
			}
		    }
		    if (j) {
			for (i=0; i<V.rkPic; i++)
			    V.loc[i] = inp[i];
			printf(
"Cone of effective curve classes is now the positive cone based at:\n");
			for (i=0; i<V.rkPic; i++)
			    printf(
"%d%c", V.loc[i], i==V.rkPic-1 ? '\n' : ' ');
		    }
		    free(inp);
		}
	    } else if (*u == '=') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    if (nknown == 0) {
			printf("No enumerative numbers known\n");
		    } else {
			p = &u[1];
			while (*p != '\0' && isspace(*p))
			    p++;
			if (*p == '\0') {
			    for (i=0; i<nknown; i++) {
				prterm(1, &vallist[i*ni], 2);
				printf(" = %s\n", qtodec(&value[i]));
			    }
			} else {
			    q = p;
			    while (*q != '\0' && !isspace(*q))
				q++;
			    *q = '\0';
			    if ((outfile = fopen(p, "w")) == NULL) {
				printf(
"error opening file %s for output\n", p);
			    } else {
				for (i=0; i<nknown; i++) {
				    fprterm(outfile, 1, &vallist[i*ni], 2);
				    fprintf(outfile,
" = %s\n", qtodec(&value[i]));
				}
				printf("Values written to file %s\n", p);
				fclose(outfile);
			    }
			}
		    }
		}
	    } else if (*u == '+') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    if (linalg && nlinear > 0) {
			printf("Linear relationships:\n");
			for (i=0; i<nlinear; i++) {
			    printf("(L%d) ", i+1);
			    prterm(1, &linlist[i*ni], 2);
			    printf(" = ...\n");
			}
		    } else {
			printf("No linear relationships\n");
		    }
		}
	    } else if (*u == '_') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    p = &u[1];
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			if (linalg && nlinear > 0) {
			    printf("Linear relationships:\n");
			    for (i=0; i<nlinear; i++) {
				printf("(L%d) ", i+1);
				prterm(1, &linlist[i*ni], 2);
				printf(" = ");
				prexpr(linrhs[i], 5);
			    }
			} else {
			    printf("No linear relationships\n");
			}
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			*q = '\0';
			i = atoi(p);
			if (i >= 1 && i <= nlinear) {
			    printf("(L%d) ", i);
			    i--;
			    prterm(1, &linlist[i*ni], 2);
			    printf(" = ");
			    prexpr(linrhs[i], 5);
			} else {
			    printf(
"Linear relations are numbered 1 through %d\n", nlinear);
			}
		    }
		}
	    } else if (*u == '-') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    p = &u[1];
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			if (linalg && nlinear > 0) {
			    printf("Linear relationships:\n");
			    for (i=0; i<nlinear; i++) {
				printf("(L%d) ", i+1);
				prterm(1, &linlist[i*ni], 2);
				printf(" = ");
				prexpr(linrhs[i], 4);
			    }
			} else {
			    printf("No linear relationships\n");
			}
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			*q = '\0';
			i = atoi(p);
			if (i >= 1 && i <= nlinear) {
			    printf("(L%d) ", i);
			    i--;
			    prterm(1, &linlist[i*ni], 2);
			    printf(" = ");
			    prexpr(linrhs[i], 3);
			} else {
			    printf(
"Linear relations are numbered 1 through %d\n", nlinear);
			}
		    }
		}
	    } else if (*u == ':') {
		printf(
"** farsta version 1.1 ** 13 May 1997 ** by Andrew Kresch\n");

		if (V.dim && restartable) {
		    printf(
"Max # associativity relations in memory:         %6d   IN USE: %6d\n",
MAXNRELS, nr);
		} else {
		    printf(
"Max # associativity relations in memory:         %6d\n", MAXNRELS);
		}

		if (V.dim && restartable) {
		    this2 = drlist[nr-1];
		    while (this2->next != NULL)
			this2 = this2->next;
		    printf(
"Max # terms in associativity relations (total):  %6d   IN USE: %6d\n",
MAXRELEXPR, this2 - drel + 1);
		} else {
		    printf(
"Max # terms in associativity relations (total):  %6d\n", MAXRELEXPR);
		}

		if (V.dim) {
		    printf(
"Max # enumerative numbers in memory:             %6d   IN USE: %6d\n",
MAXKNOWN, nknown);
		} else {
		    printf(
"Max # enumerative numbers in memory:             %6d\n", MAXKNOWN);
		}

		if (V.dim && linalg) {
		    printf(
"Max # linear relations in memory:                %6d   IN USE: %6d\n",
MAXLIN, nlinear);
		} else {
		    printf(
"Max # linear relations in memory:                %6d\n", MAXLIN);
		}

		printf(
"Max # terms in working expression:               %6d\n", MAXEXPR);

		if (V.dim && linalg && nlinear) {
		    this2 = linrhs[nlinear - 1];
		    while (this2->next != NULL)
			this2 = this2->next;
		    printf(
"Max # terms in linear relations (total):         %6d   IN USE: %6d\n",
MAXLINEXPR, this2 - linex + 1);
		} else {
		    printf(
"Max # terms in linear relations (total):         %6d\n", MAXLINEXPR);
		}

		printf(
"Space at end of linear relations buffer for new linear relation: %d\n",
LBUFFER);

		printf("Arithmetic precision: %d\n", PREC);
		printf("Verbosity: %s\n", verbose ? "yes" : "no");
		if (!linalg)
		    printf("Linear algebra: no\n");
		else if (termlim == 0)
		    printf("Linear algebra: yes  (no term limits)\n");
		else
		    printf("Linear algebra: yes  (with term limit of %d)\n",
				    termlim);
		printf("Index order: %s\n", iorder ? "reverse" : "normal");
		printf("Auto-adjust D: %s\n", autod ? "yes" : "no");
		printf("Overflow: %s\n",
			    ov ? "yes  (type \"@\" to clear)" : "no");
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    printf("Variety name: %s  (type \"A\" to show details)\n",
							    V.name);
		    printf("Number of relations: %d\n", nrels);
		    printf(
"Number of N's known: %d  (type \"=\" to display)\n", nknown);
		    if (linalg)
			printf(
"Number of linear dependencies: %d  (type \"-\" to display)\n", nlinear);
		    printf(
"\"Effective\" curve clases: the positive cone based at");
		    for (i=0; i<V.rkPic; i++)
			printf(" %d", V.loc[i]);
		    printf("\n");
		    if (restartable) {
			printf(
"Current exhaust has been interrupted; type \"X\" to restart\n");
			printf("Details of current exhaust:\n");
			cc1 = &restartbuf[0];
			cc2 = &restartbuf[rkP];
			ccc = &restartbuf[2*rkP];
			ddd = &restartbuf[3*rkP];
			printf("Equation numbers:");
			for (i=0; i<nr; ) {
			    if (i == nr-1 || rno[i+1] != rno[i]+1) {
				printf(" %d", rno[i]+1);
				i++;
			    } else {
				for (j=i+2; j<nr && rno[j] == rno[j-1]+1; j++)
				    ;
				printf(" (%d-%d)", rno[i]+1, rno[j-1]+1);
				i = j;
			    }
			}
			printf("\nLower curve class bound:");
			for (i=0; i<V.rkPic; i++)
			    printf(" %d", cc1[i]);
			printf("\nUpper curve class bound:");
			for (i=0; i<V.rkPic; i++)
			    printf(" %d", cc2[i]);
			printf("\nPoint at interruption:");
			for (i=0; i<rkP; i++)
			    printf(" %d", ccc[i]);
			printf(" ;");
			for (i=0; i<ni-rkP; i++)
			    printf(" %d", ddd[i]);
			printf("\n");
			printf("    relation (%d)   (number %d in cycle)\n",
					rno[reli]+1, reli+1);
		    }
		}
	    } else if (*u == '@') {
		if (ov) {
		    ov = 0;
		    printf("Overflow flag cleared.\n");
		}
	    } else if (*u == ';') {
		/* comment line: no operation */
	    } else if (ch == 'T') {
		time(&ti);
		times(&tis);
#ifdef CLK_TCK
		tisc = (tis.tms_utime + tis.tms_stime + tis.tms_cutime +
					tis.tms_cstime)/CLK_TCK;
#else
		tisc = (tis.tms_utime + tis.tms_stime + tis.tms_cutime +
					tis.tms_cstime)/sysconf( _SC_CLK_TCK );
#endif
		printf("CPU %5d:%.2d:%.2d   ", tisc/3600, (tisc/60)%60,
						tisc%60);
		fputs(ctime(&ti), stdout);
		fflush(stdout);
	    } else if (ch == 'F') {
		p = &u[1];
		while (*p != '\0' && isspace(*p))
		    p++;
		if (*p == '\0') {
		    printf("usage: F filename\n");
		} else {
		    q = p;
		    while (*q != '\0' && !isspace(*q))
			q++;
		    *q = '\0';
		    if ((thenewfile = fopen(p, "r")) == NULL) {
			printf("error opening file %s\n", p);
		    } else {
			printf("Reading from file %s\n", p);
			mainloop(0, 1, thenewfile);
			printf("EOF\n");
			fclose(thenewfile);
		    }
		}
	    } else if (ch == 'A') {
		p = &u[1];
		while (*p != '\0' && isspace(*p))
		    p++;
		if (*p == '\0') {
		    if (V.dim == 0) {
			printf(
"No variety specified; type A name to specify\n");
		    } else {
			printvariety(stdout, &V, 0);
		    }
		} else {
		    q = p;
		    while (*q != '\0' && !isspace(*q))
			q++;
		    *q = '\0';
		    if (V.dim)
			killvariety(&V);
		    if (scanvariety(readfile ? thefile : stdin, &V, p)) {
			printf("Error on input; type A name to try again\n");
			V.dim = 0;
		    } else {
			restartbuf = (int *)malloc(
			    (2*V.rkPic + 2*(V.rkPic + V.rkHstar - V.rkH2 - 1))*
							    sizeof(int));
			if (linalg) {   /* try to allocate linalg buffers */
			    linlist = (int *)malloc(MAXLIN*ni*sizeof(int));
			    linex = (expr *)malloc(MAXLINEXPR*sizeof(expr));
			    linex2 = (expr *)malloc(MAXLINEXPR*sizeof(expr));
			    linexd = (int *)malloc(MAXLINEXPR*ni*sizeof(int));
			    linex2d = (int *)malloc(MAXLINEXPR*ni*sizeof(int));
			    if (linlist != NULL && linex != NULL &&
					linex2 != NULL && linexd != NULL &&
					linex2d != NULL) {
				linex[0].isize = 0;
				linex[0].d = linexd;
				linex[0].next = NULL;
				linex2[0].isize = 0;
				linex2[0].d = linex2d;
				linex2[0].next = NULL;
				nlinear = 0;
			    } else {
				printf(
"Cannot allocated space for linear algebra buffers.\n");
				printf("Linear algebra flag turned off.\n");
				if (linlist)
				    free(linlist);
				if (linex)
				    free(linex);
				if (linex2)
				    free(linex2);
				if (linexd)
				    free(linexd);
				if (linex2d)
				    free(linex2d);
				linlist = NULL;
				linalg = 0;
			    }
			}
			printvariety(stdout, &V, 0);
			buildrels(&V);
			wt = V.weight;
			nrels = V.nrels;
			for (i=0; i<V.nrels; i++) {
			    if (i < 60 && !verbose) {
				printf("\n(%d)  (%d,%d,%d,%d)  / %d \\   ",
				    i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
				    V.Rl[i]+1, wt[i]);
			    } else if (i == 60 && !verbose) {
				printf(
				    "\n... (type \"G\" to get whole list)\n");
			    } else if (verbose) {
				drel[0].isize = 0;
				drel[0].d = relind;
				drel[0].next = NULL;
				if (buildtheserels(&V, drlist, drel, &i,
							    1, MAXRELEXPR))
				    printf("error building relation %d\n",i+1);
				else
				    prexpr(drel, 1);
			    }
			}
			printf("\nTotal of %d relations\n", V.nrels);
			if (autod && V.cohcodim[V.rkH2+1] != 2) {
			    printf("autod disabled\n");
			    autod = 0;
			}
		    }
		}
	    } else if (ch == 'S') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    p = &u[1];
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			outfile = stdout;
			jj = 0;
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			*q = '\0';
			if ((outfile = fopen(p, "w")) == NULL) {
			    printf("error opening file %s for output\n", p);
			    jj = -1;
			} else {
			    jj = 1;
			}
		    }
		    if (jj >= 0) {
			printvariety(outfile, &V, 1);
			if (jj == 1)
			    fclose(outfile);
		    }
		}
	    } else if (ch == 'R') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    p = &u[1];
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			printf("\nTotal of %d relations\n", nrels);
			for (i=0; i<nrels; i++) {
			    printf("\n(%d)  (%d,%d,%d,%d)  / %d \\   ",
				    i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
				    V.Rl[i]+1, wt[i]);
			    drel[0].isize = 0;
			    drel[0].d = relind;
			    drel[0].next = NULL;
			    if (buildtheserels(&V, drlist, drel, &i,
							1, MAXRELEXPR))
				printf("error building relation %d\n", i+1);
			    else
				prexpr(drel, 1);
			}
			printf("\n");
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			*q = '\0';
			i = atoi(p);
			if (i >= 1 && i <= nrels) {
			    i--;
			    printf("(%d)  (%d,%d,%d,%d)  / %d \\   ",
				    i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
				    V.Rl[i]+1, wt[i]);
			    drel[0].isize = 0;
			    drel[0].d = relind;
			    drel[0].next = NULL;
			    if (buildtheserels(&V, drlist, drel, &i,
							1, MAXRELEXPR))
				printf("error building relation %d\n", i+1);
			    else
				prexpr(drel, 1);
			} else {
			    printf("Relations are numbered 1 through %d\n",
								nrels);
			}
		    }
		}
	    } else if (ch == 'G') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    p = &u[1];
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			for (i=0; i<V.nrels; i++) {
			    printf("(%d)  (%d,%d,%d,%d)  / %d \\\n",
				    i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
				    V.Rl[i]+1, wt[i]);
			}
			printf("Total of %d relations\n", V.nrels);
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			if (*q == '\0') {
			    j = atoi(p);
			    k = 0;
			    for (i=0; i<V.nrels; i++) {
				if (wt[i] == j) {
				    printf("(%d)  (%d,%d,%d,%d)  / %d \\\n",
					i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
					V.Rl[i]+1, wt[i]);
				    k++;
				}
			    }
			    if (k)
				printf(
				"Total of %d relations with weight %d\n",
								k, j);
			    else
				printf("No relations with weight %d\n", j);
			} else {
			    *q = '\0';
			    junk[0] = atoi(p);
			    p = q + 1;
			    j = 1;
			    for (i=1; i<4; i++) {
				while (*p != '\0' && isspace(*p))
				    p++;
				if (*p == '\0') {
				    printf("not enough input\n");
				    j = 0;
				    break;
				} else {
				    q = p;
				    while (*q != '\0' && !isspace(*q))
					q++;
				    if (*q == '\0') {
					junk[i] = atoi(p);
					p = q;
				    } else {
					*q = '\0';
					junk[i] = atoi(p);
					p = q + 1;
				    }
				}
			    }
			    if (j) {
				if (junk[0] == 1 || junk[1] == 1 ||
					junk[2] == 1 || junk[3] == 1) {
				    printf(
"any relation involving identity is trivial\n");
				} else if(junk[0] < 2 || junk[0] > V.rkHstar ||
					junk[1] < 2 || junk[1] > V.rkHstar ||
					junk[2] < 2 || junk[2] > V.rkHstar ||
					junk[3] < 2 || junk[3] > V.rkHstar) {
				    printf(
"indices out of bounds; range 2 to %d\n", V.rkHstar);
				} else {
				    junk[0]--;
				    junk[1]--;
				    junk[2]--;
				    junk[3]--;
				    rnorm(junk, junk2, &k);
				    if (k == 0) {
					printf("trivial\n");
				    } else {
					j = rfind(junk2);
					if (j == 0) {
printf("error -- (%d,%d,%d,%d) should be in list but is not\n",
junk2[0], junk2[1], junk2[2], junk2[3]);
					} else {
					    printf(
					    "(%d)  (%d,%d,%d,%d)  / %d \\%s\n",
					    j, V.Ri[j-1]+1, V.Rj[j-1]+1,
					    V.Rk[j-1]+1, V.Rl[j-1]+1, wt[j-1],
					    k == -1 ? "   with opposite sign" :
									"");
					}
				    }
				}
			    }
			}
		    }
		}
	    } else if (ch == 'B') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    p = &u[1];
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			printf("Total of %d relations\n", nrels);
			j = checkclassassoc(&V, 0, nrels-1, 0);
#if 0
			j = 0;
			for (i=0; i<nrels; i++) {
			    if (checkclassassoc(&V, i)) {
				printf(
"(%d)  (%d,%d,%d,%d)  / %d \\    associativity fails\n",
					i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
					V.Rl[i]+1, wt[i]);
				j = 1;
			    }
			}
#endif
			if (j == 0)
			    printf("Associativity checks for all of them\n");
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			*q = '\0';
			i = atoi(p);
			if (i >= 1 && i <= nrels) {
			    i--;
			    checkclassassoc(&V, i, i, 1);
#if 0
			    j = checkclassassoc(&V, i);
			    printf(
"(%d)  (%d,%d,%d,%d)  / %d \\    associativity %s\n",
				i+1, V.Ri[i]+1, V.Rj[i]+1, V.Rk[i]+1,
				V.Rl[i]+1, wt[i], j ? "fails" : "checks");
#endif
			} else {
			    printf("Relations are numbered 1 through %d\n",
								nrels);
			}
		    }
		}
	    } else if (ch == 'D') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    autod = atoi(&u[1]);
		    if (autod) {
			if (V.cohcodim[V.rkH2+1] == 2)
			    printf("Automatically adjusts d_%s.\n",
						V.cohname[V.rkH2+1]);
			else
			    printf(
"Sorry, requires %s to be a codimension 2 class.\n", V.cohname[V.rkH2+1]);
		    } else {
			printf("Won't automatically adjusts d_%s.\n",
						    V.cohname[V.rkH2+1]);
		    }
		}
	    } else if (ch == 'Q') {
		exit(0);
	    } else if (ch == 'C') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    nknown = 0;
		    nlinear = 0;
		    printf("Previous enumerative knowledge cleared\n");
		}
	    } else if (ch == 'V') {
		verbose = atoi(&u[1]);
		if (verbose)
		    printf("Now noisy.\n");
		else
		    printf("Now quiet.\n");
	    } else if (ch == 'O') {
		iorder = atoi(&u[1]);
		if (iorder)
		    printf("Exhaust with indices in reverse order.\n");
		else
		    printf("Exhaust with indices in default order.\n");
	    } else if (ch == 'L') {
		p = &u[1];
		while (*p != '\0' && isspace(*p))
		    p++;
		if (*p == '-') {
		    if (linalg == 0) {
			printf("We're not using linear algebra; no effect\n");
		    } else if (nlinear == 0) {
			printf("No linear relationships; no effect\n");
		    } else {
			p++;
			if (*p == '\0') {
			    k = 1;
			} else {
			    q = p;
			    while (*q != '\0' && !isspace(*q))
				q++;
			    if (*q == '\0') {
				k = atoi(p);
			    } else {
				*q = '\0';
				k = atoi(p);
				p = q + 1;
				while (*p != '\0' && isspace(*p))
				    p++;
				if (*p == '\0')
				    termlim = 0;
				else
				    termlim = atoi(p);
			    }
			}
			nlinear -= k;
			if (nlinear >= 0) {
			    if (k == 1)
				printf("One linear relationship removed.\n");
			    else
				printf("%d linear relationships removed.\n",k);
			} else {
			    nlinear = 0;
			    printf("All linear relationships removed.\n");
			}
		    }
		} else {
		    if (*p == '\0') {
			if (linalg) {
			    if (termlim) {
				printf(
"Currently using linear algebra, with term limit of %d terms.\n", termlim);
			    } else {
				printf(
"Currently using linear algebra, with no term limit.\n");
			    }
			} else {
			    printf("Currently not using linear algebra.\n");
			}
		    } else {
			otermlim = termlim;
			termlim = 0;
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			if (*q == '\0') {
			    k = atoi(p);
			} else {
			    *q = '\0';
			    k = atoi(p);
			    p = q + 1;
			    while (*p != '\0' && isspace(*p))
				p++;
			    if (*p == '\0')
				termlim = 0;
			    else
				termlim = atoi(p);
			}
			if (k) {
			    if (linalg) {
				if (termlim) {
				    if (termlim == otermlim) {
					printf(
"Already using linear algebra with limit of %d terms.\n", termlim);
				    } else {
					printf(
"Already using linear algebra; now with limit of %d terms.\n", termlim);
				    }
				} else {
				    if (otermlim) {
					printf(
"Already using linear algebra; now with no term limit.\n");
				    } else {
					printf(
"Already using linear algebra.\n");
				    }
				}
				if (nlinear > 0)
				    printf(
"Enter the command L0 followed by L1 to clear all linear relationships.\n");
			    } else {
				linalg = 1;
				if (V.dim) {
				    if (linlist) {
					free(linlist);
					free(linex);
					free(linex2);
					free(linexd);
					free(linex2d);
				    }
				    linlist = (int *)malloc(MAXLIN*ni*
							sizeof(int));
				    linex = (expr *)malloc(MAXLINEXPR*
							sizeof(expr));
				    linex2 = (expr *)malloc(MAXLINEXPR*
							sizeof(expr));
				    linexd = (int *)malloc(MAXLINEXPR*ni*
							sizeof(int));
				    linex2d = (int *)malloc(MAXLINEXPR*ni*
							sizeof(int));
				    if (linlist != NULL && linex != NULL &&
						linex2 != NULL &&
						linexd != NULL &&
						linex2d != NULL) {
					linex[0].isize = 0;
					linex[0].d = linexd;
					linex[0].next = NULL;
					linex2[0].isize = 0;
					linex2[0].d = linex2d;
					linex2[0].next = NULL;
					nlinear = 0;
					printf(
"Will now use linear algebra to obtain more numbers.\n");
					if (termlim)
					    printf(
"Limit on linear relationships of %d terms.\n", termlim);
				    } else {
					printf(
"Cannot allocated space for linear algebra buffers.\n");
					printf(
"Linear algebra flag turned off.\n");
					if (linlist)
					    free(linlist);
					if (linex)
					    free(linex);
					if (linex2)
					    free(linex2);
					if (linexd)
					    free(linexd);
					if (linex2d)
					    free(linex2d);
					linlist = NULL;
					linalg = 0;
				    }
				} else {
				    printf(
"Will now use linear algebra to obtain more numbers.\n");
				    if (termlim)
					printf(
"Limit on linear relationships of %d terms.\n", termlim);
				}
			    }
			} else {
			    linalg = 0;
			    printf("Won't use linear algebra.\n");
			    if (linlist) {
				free(linlist);
				free(linex);
				free(linex2);
				free(linexd);
				free(linex2d);
				linlist = NULL;
			    }
			}
		    }
		}
	    } else if (ch == 'N') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    inp = (int *)malloc(ni*sizeof(int));
		    p = &u[1];
		    j = 1;
		    l = 1;
		    for (i=0; i<1+ni; i++) {
			while (*p != '\0' && (isspace(*p) ||
				(i<ni && *p == '(') || *p == ')' ||
				*p == ',' || *p == ';' || *p == '=') )
			    p++;
			if (*p == '\0') {
			    if (i == ni) {
				l = 2;
			    } else {
				printf("not enough input\n");
				j = 0;
			    }
			    break;
			} else {
			    if (i < ni) {
				q = p;
				while (*q != '\0' && !(isspace(*q) ||
					*q == '(' || *q == ')' || *q == ',' ||
					*q == ';' || *q == '=') )
				    q++;
				if (*q == '\0') {
				    inp[i] = atoi(p);
				    p = q;
				} else {
				    *q = '\0';
				    inp[i] = atoi(p);
				    p = q + 1;
				}
			    } else {
				if (p[0] == '?') {
				    l = 0;
				} else {
				    dectoq(p, &z, NULL);
				}
			    }
			}
		    }
		    if (j) {
			k = 0;
			for (i=0; i<ni && !k; i++)
			    if (inp[i] < ((i<V.rkPic)?V.loc[i]:0))
				k = 1;
			if (k) {
			    printf(
"curve classes must be in positive cone; cohomology class multiplicities \
must\nbe nonnegative\n");
			} else {
			    ccc = &inp[0];
			    ddd = &inp[rkP];
			    k = 0;
			    for (i=0; i<rkP; i++)
				k += ccc[i]*V.c1t[i];
			    for (i=1; i<ni-rkP; i++)
				k -= ddd[i]*(V.cohcodim[1+V.rkH2+i] - 1);
			    k += V.dim - 3;
			    if (k == ddd[0]*(V.cohcodim[1+V.rkH2] - 1)) {
				j = 1;
			    } else {
				if (autod) {
				    printf("autod correction...");
				    ddd[0] = k;
				    for (i=0; i<ni; i++)
					printf(" %d", inp[i]);
				    printf("\n");
				    if (k >= 0) {
					j = 1;
				    } else {
					printf("this is disallowed\n");
					j = 0;
				    }
				} else {
				    printf(
"dimension is wrong -- not allowed\n");
				    j = 0;
				}
			    }
			    if (j) {
				k = dowehave(inp, &b);
				if (l == 2) {
				    prterm(1, inp, 2);
				    printf(" = %s\n", k ? qtodec(&b) : "?");
				} else if (l == 1) {
				    if (k) {
					printf("Previous: ");
					prterm(1, inp, 2);
					printf(" = %s\n", qtodec(&b));
					removeknowledge(inp);
					addknowledge(inp, &z);
					printf("Now: ");
					prterm(1, inp, 2);
					printf(" = %s\n", qtodec(&z));
					if (linalg) {
					    printf(
"Warning: old value may have been incorporated into linear relationships\n");
					    printf(
"Enter the command L0 followed by L1 to clear all linear relationships\n");
					}
				    } else {
					if (linalg) {
					    k = doreln(&V, 0, NULL, NULL,
						    inp, &z, 1, 0, verbose,
						    1, &nadded);
					    switch(k) {
					    case 9:
						printf(
"Error: overflow of linear relationships buffer.\n");
						printf(
"Try removing some relationships with \"L-\", then try again.\n");
						break;
					    case -1:
					    case -2:
					    case -4:
						printf(
"Error: %s\n", k==-1 ? "expressions got to long, had to truncate" :
k==-2 ? "overflow" : "critical overflow");
						printf("Setting by hand: ");
						addknowledge(inp, &z);
						prterm(1, inp, 2);
						printf(" = %s\n", qtodec(&z));
						printf(
"This change may not be reflected in the linear relationships.\n");
						break;
					    case -3:
						printf("User interrupt\n");
						break;
					    case 10:
					    case 11:
						printf("Gives ");
						for (i=0; i<nadded; i++) {
						    if (i>0)
							printf("        ");
						    prterm(1,
			    &vallist[(nknown-nadded+i)*ni], 2);
						    printf(" = %s\n",
			    qtodec(&value[nknown-nadded+i]));
						}
						break;
					    default:
						printf(
"unexpected return code %d\n", k);
						break;
					    }
					} else {
					    addknowledge(inp, &z);
					    prterm(1, inp, 2);
					    printf(" = %s\n", qtodec(&z));
					}
				    }
				} else {
				    removeknowledge(inp);
				    prterm(1, inp, 2);
				    printf(" = ?\n");
				    if (k && linalg) {
					printf(
"Warning: old value may have been incorporated into linear relationships.\n");
					printf(
"Enter the command L0 followed by L1 to clear all linear relationships\n");
				    }
				}
			    }
			}
		    }
		    free(inp);
		}
	    } else if (ch == 'E') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    inp = (int *)malloc((1+2*ni)*sizeof(int));
		    outp = inp + 1 + ni;
		    p2 = &u[1];
		    j = 1;
		    while (*p2 != '\0' && isspace(*p2))
			p2++;
		    if (*p2 == '(') {
			p = p2;
			while (*p != '\0' && *p != ')')
			    p++;
			if (*p == '\0') {
			    printf("No closing parentheses\n");
			    j = 0;
			} else {
			    p++;
			    while (*p != '\0' && isspace(*p))
				p++;
			}
		    } else {
			p = p2;
			if (*p == '\0') {
			    printf("not enough input\n");
			    j = 0;
			} else {
			    q = p;
			    while (*q != '\0' && !isspace(*q))
				q++;
			    if (*q == '\0') {
				inp[0] = atoi(p);
				p = q;
			    } else {
				*q = '\0';
				inp[0] = atoi(p);
				p = q + 1;
			    }
			}
		    }
		    if (j) {
			q = p;
			for (i=1; i<1+ni; i++) {
			    while (*p != '\0' && isspace(*p))
				p++;
			    if (*p == '\0') {
				printf("not enough input\n");
				j = 0;
				break;
			    } else {
				q = p;
				while (*q != '\0' && !isspace(*q))
				    q++;
				if (*q == '\0') {
				    inp[i] = atoi(p);
				    p = q;
				} else {
				    *q = '\0';
				    inp[i] = atoi(p);
				    p = q + 1;
				}
			    }
			}
		    }
		    if (j) {
			for (i=1; i<1+ni && j; i++) {
			    if (inp[i] < ((i<1+V.rkPic) ? V.loc[i-1] : 0)) {
				j = 0;
				if (i<1+V.rkPic)
				    printf(
"curve class must be in positive cone\n");
				else
				    printf(
"cohomology class multiplicities must be nonnegative\n");
			    }
			}
		    }
		    if (*p2 == '(') {
			p = p2;
			if (j) {
			    p++;
			    while (*p != '\0' && isspace(*p))
				p++;
			    while (*p != '\0' && *p != ')') {
				q = p;
				while (*q != '\0' && !isspace(*q) && *q != ')'
							    && *q != '-')
				    q++;
				if (*q == '\0') {
				    kk = atoi(p);
				    p = q;
				} else {
				    ss = *q;
				    *q = '\0';
				    kk = atoi(p);
				    *q = ss;
				    p = q;
				}
				while (*p != '\0' && isspace(*p))
				    p++;
				if (*p != '-') {
				    if (kk < 1 || kk > nrels) {
					printf(
"relation number out of bounds: range is [1,%d]\n", nrels);
				    } else {
					inp[0] = kk;


					ccc = &inp[1];
					ddd = &inp[1+rkP];
					k = 0;
					whichrel = inp[0] - 1;
					for (i=0; i<rkP; i++)
					    k += ccc[i]*V.c1t[i];
					for (i=1; i<ni-rkP; i++)
					    k -= ddd[i]*(V.cohcodim[1+V.rkH2+i]
								- 1);
					k -= wt[whichrel];
					if (k == ddd[0]*(V.cohcodim[1+V.rkH2]
								- 1)) {
					    j = 1;
					} else { /* no autod correction here */
					    j = 0; /* no error message either*/
					} /* we could be doing lots of these */
					if (j) {
					    drel[0].isize = 0;
					    drel[0].d = relind;
					    drel[0].next = NULL;
					    if (buildtheserels(&V, drlist,
					    drel, &whichrel, 1, MAXRELEXPR)) {
						printf(
"error building relation %d\n", whichrel+1);
					    } else {
						k = doreln(&V, drel, ccc,
	ddd, outp, &z, linalg, termlim, verbose, 0, &nadded);
						if (verbose || (k != 0 &&
	k != 1 && k != 2 && k != 8)) {
						    printf("Relation (%d)  ",
								whichrel+1);
						    for (i=0; i<rkP; i++)
							printf(" %d", ccc[i]);
						    printf(" ;");
						    for (i=0; i<ni-rkP; i++)
							printf(" %d", ddd[i]);
						}
						switch(k) {
						case -1:
						    printf(
"  gives error: truncation\n");
						    break;
						case -2:
						    printf(
"  gives error: overflow\n");
						    break;
						case -3:
						    printf(
"  gives error: user interrupt\n");
						    if (readfile) {
							printf(
"You may view data or make changes\n");
							printf(
"To resume this exhaust, reenter:\n  \"E\" (remaining relations)");
							for (i=0; i<rkP; i++)
							    printf(" %d",
								    ccc[i]);
							printf(" ");
							for (i=0; i<ni-rkP;i++)
							    printf(" %d",
								    ddd[i]);
							printf(
"\nEnter \"W\" to resume from file, or \"Z\" to close file and proceed.\n");
							if (mainloop(0, 0,
								    stdin)==2)
							    return 0;
						    } else {
							printf(
"You may view data or make changes\n");
							printf(
"To resume this exhaust, reenter:\n  \"E\" (remaining relations)");
							for (i=0; i<rkP; i++)
							    printf(" %d",
								    ccc[i]);
							printf(" ");
							for (i=0; i<ni-rkP;i++)
							    printf(" %d",
								    ddd[i]);
							printf("\n");
						    }
						    goto eexhout;
						case -4:
						    printf(
"  gives error: critical overflow -- knowledge of linear relations cleared\n");
						    break;
						case 0:
						    if (verbose)
							printf(
"  is trivial\n");
						    break;
						case 1:
						    if (verbose)
							printf(
"  is complicated -- no checks\n");
						    break;
						case 2:
						    if (verbose)
							printf("  checks\n");
						    break;
						case 3:
						    printf("  FAILS!!!!\n");
						    break;
						case 8:
						    if (verbose)
							printf(
"  gives linear relationship, discarded due to term limit\n");
						    break;
						case 5:
						case 10:
						case 11:
#if 0
						    printf("  gives ");
						    prterm(1, outp, 2);
						    printf(" = %s\n",
								qtodec(&z));
#endif
						    printf("  gives ");
						    if (k == 10)
							printf(
"new linear relationship -- now have %d -- and\n", nlinear);
						    for (i=0; i<nadded; i++) {
							if (i>0 || k == 10)
							    printf(
"                                   ");
							prterm(1,
			    &vallist[(nknown-nadded+i)*ni], 2);
							printf(
" = %s\n", qtodec(&value[nknown-nadded+i]));
						    }
						    break;
						case 9:
						    printf(
"  yields out-of-memory during linalg\n");
						    break;
						case 12:
						    printf(
"  gives new linear relationship -- now have %d\n", nlinear);
						    break;
						}
					    }
					}




				    }
				} else {
				    p++;
				    while (*p != '\0' && isspace(*p))
					p++;
				    q = p;
				    while (*q != '\0' && !isspace(*q) &&
								    *q != ')')
					q++;
				    if (*q == '\0') {
					ll = atoi(p);
					p = q;
				    } else {
					ss = *q;
					*q = '\0';
					ll = atoi(p);
					*q = ss;
					p = q;
				    }
				    if (kk < 1 || ll > nrels || kk > ll) {
					printf(
"range of relations should be an interval within [1,%d]\n", nrels);
				    } else {
					for (l=kk; l<=ll; l++) {
					    inp[0] = l;


					    ccc = &inp[1];
					    ddd = &inp[1+rkP];
					    k = 0;
					    whichrel = inp[0] - 1;
					    for (i=0; i<rkP; i++)
						k += ccc[i]*V.c1t[i];
					    for (i=1; i<ni-rkP; i++)
						k -= ddd[i]*(V.cohcodim[1+
							    V.rkH2+i] - 1);
					    k -= wt[whichrel];
					    if (k == ddd[0]*(V.cohcodim[1+
							    V.rkH2] - 1)) {
						j = 1;
					    } else { /* no autod corr here */
						j = 0; /* no error msg either*/
					    } /* could be doing lots of these*/
					    if (j) {
						drel[0].isize = 0;
						drel[0].d = relind;
						drel[0].next = NULL;
						if (buildtheserels(&V, drlist,
			    drel, &whichrel, 1, MAXRELEXPR)) {
						    printf(
"error building relation %d\n", whichrel+1);
						} else {
						    k = doreln(&V, drel,
	ccc, ddd, outp, &z, linalg, termlim, verbose, 0, &nadded);
						    if (verbose || (k != 0 &&
	k != 1 && k != 2 && k != 8)) {
							printf(
"Relation (%d)  ", whichrel+1);
							for (i=0; i<rkP; i++)
							    printf(" %d",
								    ccc[i]);
							printf(" ;");
							for (i=0; i<ni-rkP;i++)
							    printf(" %d",
								    ddd[i]);
						    }
						    switch(k) {
						    case -1:
							printf(
"  gives error: truncation\n");
							break;
						    case -2:
							printf(
"  gives error: overflow\n");
							break;
						    case -3:
							printf(
"  gives error: user interrupt\n");
							if (readfile) {
							    printf(
"You may view data or make changes\n");
							    printf(
"To resume this exhaust, reenter:\n  \"E\" (remaining relations)");
							    for (i=0; i<rkP;
									i++)
								printf(" %d",
								    ccc[i]);
							    printf(" ");
							    for (i=0; i<ni-rkP;
									i++)
								printf(" %d",
								    ddd[i]);
							    printf(
"\nEnter \"W\" to resume from file, or \"Z\" to close file and proceed.\n");
							    if (mainloop(0, 0,
								    stdin)==2)
								return 0;
							} else {
							    printf(
"You may view data or make changes\n");
							    printf(
"To resume this exhaust, reenter:\n  \"E\" (remaining relations)");
							    for (i=0; i<rkP;
									i++)
								printf(" %d",
								    ccc[i]);
							    printf(" ");
							    for (i=0; i<ni-rkP;
									i++)
								printf(" %d",
								    ddd[i]);
							    printf("\n");
							}
							goto eexhout;
						    case -4:
							printf(
"  gives error: critical overflow -- knowledge of linear relations cleared\n");
							break;
						    case 0:
							if (verbose)
							    printf(
"  is trivial\n");
							break;
						    case 1:
							if (verbose)
							    printf(
"  is complicated -- no checks\n");
							break;
						    case 2:
							if (verbose)
							    printf(
"  checks\n");
							break;
						    case 3:
							printf(
"  FAILS!!!!\n");
							break;
						    case 8:
							if (verbose)
							    printf(
"  gives linear relationship, discarded due to term limit\n");
							break;
						    case 5:
						    case 10:
						    case 11:
#if 0
							printf("  gives ");
							prterm(1, outp, 2);
							printf(" = %s\n",
								qtodec(&z));
#endif
							printf("  gives ");
							if (k == 10)
							    printf(
"new linear relationship -- now have %d -- and\n", nlinear);
							for (i=0; i<nadded;
									i++) {
							    if (i>0 || k == 10)
								printf(
"                                   ");
							    prterm(1,
		&vallist[(nknown-nadded+i)*ni], 2);
							    printf(
" = %s\n", qtodec(&value[nknown-nadded+i]));
							}
							break;
						    case 9:
							printf(
"  yields out-of-memory during linalg\n");
							break;
						    case 12:
							printf(
"  gives new linear relationship -- now have %d\n", nlinear);
							break;
						    }
						}
					    }



					}
				    }
				    while (*p != '\0' && isspace(*p))
					p++;
				}
			    }
eexhout:
			    ;
			}
		    } else {
			if (j) {
			    if (inp[0] < 1 || inp[0] > nrels) {
				printf("relations are numbered 1 through %d",
							nrels);
			    } else {
				ccc = &inp[1];
				ddd = &inp[1+rkP];
				k = 0;
				whichrel = inp[0] - 1;
				for (i=0; i<rkP; i++)
				    k += ccc[i]*V.c1t[i];
				for (i=1; i<ni-rkP; i++)
				    k -= ddd[i]*(V.cohcodim[1+V.rkH2+i] - 1);
				k -= wt[whichrel];
				if (k == ddd[0]*(V.cohcodim[1+V.rkH2] - 1)) {
				    j = 1;
				} else {
				    if (autod) {
					printf("autod correction...");
					ddd[0] = k;
					for (i=0; i<ni; i++)
					    printf(" %d", inp[i+1]);
					printf("\n");
					if (k >= 0) {
					    j = 1;
					} else {
					    printf("this is disallowed\n");
					    j = 0;
					}
				    } else {
					printf(
"dimension is wrong -- trivial\n");
					j = 0;
				    }
				}
				if (j) {
				    drel[0].isize = 0;
				    drel[0].d = relind;
				    drel[0].next = NULL;
				    if (buildtheserels(&V, drlist, drel,
						&whichrel, 1, MAXRELEXPR)) {
					printf("error building relation %d\n",
						    whichrel+1);
				    } else {
					k = doreln(&V, drel, ccc, ddd,
		    outp, &z, linalg, termlim, verbose, 0, &nadded);
					printf("Relation (%d)  ", whichrel+1);
					for (i=0; i<rkP; i++)
					    printf(" %d", ccc[i]);
					printf(" ;");
					for (i=0; i<ni-rkP; i++)
					    printf(" %d", ddd[i]);
					switch(k) {
					case -1:
					    printf(
"  gives error: truncation\n");
					    break;
					case -2:
					    printf(
"  gives error: overflow\n");
					    break;
					case -3:
					    printf(
"  gives error: user interrupt\n");
					    break;
					case -4:
					    printf(
"  gives error: critical overflow -- knowledge of linear relations cleared\n");
					    break;
					case 0:
					    printf(
"  is trivial\n");
					    break;
					case 1:
					    printf(
"  is complicated -- no checks\n");
					    break;
					case 2:
					    printf("  checks\n");
					    break;
					case 3:
					    printf("  FAILS!!!!\n");
					    break;
					case 8:
					    printf(
"  gives linear relationship, discarded due to term limit\n");
					    break;
					case 5:
					case 10:
					case 11:
#if 0
					    printf("  gives ");
					    prterm(1, outp, 2);
					    printf(" = %s\n", qtodec(&z));
#endif
					    printf("  gives ");
					    if (k == 10)
						printf(
"new linear relationship -- now have %d -- and\n", nlinear);
					    for (i=0; i<nadded; i++) {
						if (i>0 || k == 10)
						    printf(
"                                   ");
						prterm(1,
			    &vallist[(nknown-nadded+i)*ni], 2);
						printf(" = %s\n",
			    qtodec(&value[nknown-nadded+i]));
					    }
					    break;
					case 9:
					    printf(
"  yields out-of-memory during linalg\n");
					    break;
					case 12:
					    printf(
"  gives new linear relationship -- now have %d\n", nlinear);
					    break;
					}
				    }
				}
			    }
			}
		    }
		    free(inp);
		}
	    } else if (ch == 'X') {
		if (V.dim == 0) {
		    printf("No variety specified\n");
		} else {
		    cc1 = &restartbuf[0];
		    cc2 = &restartbuf[rkP];
		    ccc = &restartbuf[2*rkP];
		    ddd = &restartbuf[3*rkP];
		    outp = &restartbuf[2*rkP+ni];
		    p = &u[1];

		    jj = 1;
		    while (*p != '\0' && isspace(*p))
			p++;
		    if (*p == '\0') {
			if (restartable) {
			    jj = 2;
			} else {
			    printf("not enough input\n");
			    jj = 0;
			}
		    } else if (*p == '(') {
			nr = 0;
			p++;
			while (*p != '\0' && isspace(*p))
			    p++;
			while (*p != '\0' && *p != ')') {
			    q = p;
			    while (*q != '\0' && !isspace(*q) &&
						*q != ')' && *q != '-')
				q++;
			    if (*q == '\0') {
				k = atoi(p);
				p = q;
			    } else {
				ss = *q;
				*q = '\0';
				k = atoi(p);
				*q = ss;
				p = q;
			    }
			    while (*p != '\0' && isspace(*p))
				p++;
			    if (*p != '-') {
				if (k < 1 || k > nrels) {
				    printf(
"relation number out of bounds: range is [1,%d]\n", nrels);
				    jj = 0;
				    break;
				}
				if (nr == MAXNRELS) {
				    printf(
"maximum number of relations allowed is %d\n", MAXNRELS);
				    jj = 0;
				    break;
				} else {
				    rno[nr++] = k-1;
				}
			    } else {
				p++;
				while (*p != '\0' && isspace(*p))
				    p++;
				q = p;
				while (*q != '\0' && !isspace(*q) && *q != ')')
				    q++;
				if (*q == '\0') {
				    l = atoi(p);
				    p = q;
				} else {
				    ss = *q;
				    *q = '\0';
				    l = atoi(p);
				    *q = ss;
				    p = q;
				}
				if (k < 1 || l > nrels || k > l ||
						    l - k + 1 > MAXNRELS) {
				    printf(
"range of relations should be an interval within [1,%d]\n", nrels);
				    jj = 0;
				    break;
				} else if (nr + l - k + 1 > MAXNRELS) {
				    printf(
"maximum number of relations allowed is %d\n", MAXNRELS);
				    jj = 0;
				    break;
				} else {
				    for (i=0; i<l-k+1; i++) {
					rno[nr++] = k - 1 + i;
				    }
				}
				while (*p != '\0' && isspace(*p))
				    p++;
			    }
			}
			if (jj == 0) {
			} else if (*p == ')') {
			    p++;
			    if (nr == 0) {
				printf(
"please specify relations to exhaust with\n");
				jj = 0;
			    }
			} else {
			    printf("no closing parenthesis\n");
			    jj = 0;
			}
		    } else {
			q = p;
			while (*q != '\0' && !isspace(*q))
			    q++;
			if (*q == '\0') {
			    k = atoi(p);
			    p = q;
			} else {
			    *q = '\0';
			    k = atoi(p);
			    p = q + 1;
			}
			while (*p != '\0' && isspace(*p))
			    p++;
			if (*p == '\0') {
			    printf("not enough input\n");
			    jj = 0;
			} else {
			    q = p;
			    while (*q != '\0' && !isspace(*q))
				q++;
			    if (*q == '\0') {
				l = atoi(p);
				p = q;
			    } else {
				*q = '\0';
				l = atoi(p);
				p = q + 1;
			    }
			    if (k < 1 || l > nrels || k > l ||
						l - k + 1 > MAXNRELS) {
				printf(
"first two entries must be a range of relations, within [1,%d]\n", nrels);
				printf("maximum number: %d\n", MAXNRELS);
				jj = 0;
			    } else {
				for (i=0; i<l-k+1; i++) {
				    rno[i] = k - 1 + i;
				}
				nr = l-k+1;
			    }
			}
		    }
		    if (jj) {
			if (jj == 1) {
			    for (i=0; i<2*rkP; i++) {
				while (*p != '\0' && isspace(*p))
				    p++;
				if (*p == '\0') {
				    printf("not enough input\n");
				    jj = 0;
				    break;
				} else {
				    q = p;
				    while (*q != '\0' && !isspace(*q))
					q++;
				    if (*q == '\0') {
					restartbuf[i] = atoi(p);
					p = q;
				    } else {
					*q = '\0';
					restartbuf[i] = atoi(p);
					p = q + 1;
				    }
				}
			    }
			}
		    }
		    if (jj) {
			k = 0;
			for (i=0; i<rkP && !k; i++)
			    if (restartbuf[i] < V.loc[i] ||
				    restartbuf[i+rkP] < V.loc[i] ||
				    restartbuf[i] > restartbuf[i+rkP])
				k = 1;
			if (k) {
			    printf(
"curve class indices must be a range in the effective cone\n");
			} else {
			    drel[0].isize = 0;
			    drel[0].d = relind;
			    drel[0].next = NULL;
			    if (buildtheserels(&V, drlist, drel, rno,
						nr, MAXRELEXPR)) {
				printf(
"Out of space building relations, try reducing number\n");
			    } else {
				if (jj == 1)
				    for (i=0; i<rkP; i++)
					ccc[i] = cc1[i];
				while (1) {
				    cdota = 0;
				    for (i=0; i<rkP; i++)
					cdota += ccc[i]*V.c1t[i];
				    if (jj == 1) {
					for (i=1; i<ni-rkP; i++)
					    ddd[i] = 0;
				    }
				    while (1) {
					m = cdota;
					for (j=1; j<ni-rkP; j++)
					    m -= ddd[j]*(V.cohcodim[j+
								V.rkH2+1]-1);
					if (jj == 1)
					    reli = 0;
					jj = 1;
					for (; reli < nr; reli++) {
					    whichrel = rno[reli];
					    if (m >= wt[whichrel] &&
			    (m-wt[whichrel])%(V.cohcodim[V.rkH2+1]-1)==0) {
						ddd[0] =
			    (m-wt[whichrel])/(V.cohcodim[V.rkH2+1]-1);
						k = doreln(&V, drlist[reli],
	    ccc, ddd, outp, &z, linalg, termlim, verbose, 0, &nadded);
						if (verbose ||
	    (k != 0 && k != 1 && k != 2 && k != 8)) {
						    printf("Relation (%d)  ",
								whichrel+1);
						    for (i=0; i<rkP; i++)
							printf(" %d", ccc[i]);
						    printf(" ;");
						    for (i=0; i<ni-rkP; i++)
							printf(" %d", ddd[i]);
						}
						switch(k) {
						case -1:
						    printf(
"  gives error: truncation\n");
						    break;
						case -2:
						    printf(
"  gives error: overflow\n");
						    break;
						case -3:
						    printf(
"  gives error: user interrupt\n");
						    if (readfile) {
							printf(
"You may view data or make changes\n");
							printf(
"To continue executing commands from file, use:\n");
							printf(
"  \"W\" to resume with next command (current exhaust will be killed);\n");
							printf(
"  \"X\" to resume exhaust then return control to user;\n");
							printf(
"  \"Y\" to resume exhaust then continue executing commands from file;\n");
							printf(
"  \"Z\" to close input file and proceed.\n");
							restartable = 1;
							if ((i=mainloop(0, 0,
								stdin)) == 0)
							    goto quickout;
							else if (i == 2)
							    return 0;
						    } else {
							printf(
"You may view data or make changes, and restart by typing just \"X\"\n");
							restartable = 1;
							goto quickout;
						    }
						    break;
						case -4:
						    printf(
"  gives error: critical overflow -- knowledge of linear relations cleared\n");
						    break;
						case 0:
						    if (verbose)
							printf(
"  is trivial\n");
						    break;
						case 1:
						    if (verbose)
							printf(
"  is complicated -- no checks\n");
						    break;
						case 2:
						    if (verbose)
							printf("  checks\n");
						    break;
						case 3:
						    printf("  FAILS!!!!\n");
						    break;
						case 8:
						    if (verbose)
							printf(
"  gives linear relationship, discarded due to term limit\n");
						    break;
						case 5:
						case 10:
						case 11:
						    printf("  gives ");
						    if (k == 10)
							printf(
"new linear relationship -- now have %d -- and\n", nlinear);
						    for (i=0; i<nadded; i++) {
							if (i>0 || k == 10)
							    printf(
"                                   ");
							prterm(1,
		    &vallist[(nknown-nadded+i)*ni], 2);
							printf(" = %s\n",
		    qtodec(&value[nknown-nadded+i]));
						    }
#if 0
						    prterm(1, outp, 2);
						    printf(" = %s\n",
								qtodec(&z));
#endif
						    break;
						case 9:
						    printf(
"  yields out-of-memory during linalg\n");
						    break;
						case 12:
						    printf(
"  gives new linear relationship -- now have %d\n", nlinear);
						    break;
						}
					    }
					}

					if (iorder == 0) {
					    for (i=1; i<ni-rkP; i++) {
						m = 0;
						for (j=i; j<ni-rkP; j++)
						    m += ddd[j]*(V.cohcodim[j+
								V.rkH2+1]-1);
						if (m <= cdota)
						    break;
					    }
					    if (i == ni-rkP)
						break;
					    ddd[i]++;
					    while (--i >= 0)
						ddd[i] = 0;
					} else {
					    for (i=ni-rkP-1; i>=1; i--) {
						m = 0;
						for (j=i; j>=1; j--)
						    m += ddd[j]*(V.cohcodim[j+
								V.rkH2+1]-1);
						if (m <= cdota)
						    break;
					    }
					    if (i == 0)
						break;
					    ddd[i]++;
					    while (++i < ni-rkP)
						ddd[i] = 0;
					}
				    }
				    for (i=rkP-1; i>=0 && ccc[i] == cc2[i];
									i--)
					;
				    if (i<0)
					break;
				    ccc[i]++;
				    for (j=i+1; j<rkP; j++)
					ccc[j] = cc1[j];
				}
				restartable = 0;
quickout:                           ;
			    }
			}
		    }
		}
	    } else if (ch == 'W') {
		if (V.dim && !first) {
		    return 0;
		} else {
		    printf(
"Only valid during break to user while reading commands from file\n");
		}
	    } else if (ch == 'Y') {
		if (V.dim && restartable && !first) {
		    return 1;
		} else {
		    printf(
"Only valid during interruped exhaust while reading commands from file\n");
		}
	    } else if (ch == 'Z') {
		if (V.dim && !first) {
		    return 2;
		} else {
		    printf(
"Only valid during break to user while reading commands from file\n");
		}
	    } else {
		if (*u != '?')
		    printf("Command \"%c\" unrecognized\n",
				isalpha(*u) ? ch : *u);
		printf(
"Commands:    (may be uppercase or lowercase)\n");
		printf(
"             brackets indicate entry which may be optionally specified\n");
		printf(
"F file           read commands from file\n");
		printf(
"A name           specify new variety\n");
		printf(
"A                display current variety\n");
		printf(
"B [n]            check classical associativity (wrt all or just \
indicated reln)\n");
		printf(
"S [outfile]      save variety to outfile (or screen)\n");
		printf(
"C                clear enumerative knowledge\n");
		printf(
"R [n]            display relation n or all if none specified\n");
		printf(
"G                show whole list of relation numbers, classes, and \
weights\n");
		printf(
"G a b c d\n");
		printf(
"G w              show number of relation(s) with given indexing or weight\n");
		printf(
"Vn               set verbosity (1=verbose  0=quiet)\n");
		printf(
"On               set order for exhaust (1=reverse on indices  0=normal)\n");
		printf(
"Dn               set auto-adjust first d (1=yes 0=no)\n");
		printf(
"Ln [termlimit]   set linear algebra usage (1=yes 0=no)\n");
		printf(
"L-[n]            remove n linear relations from end of list (default: 1)\n");
		printf(
"N c .. d ..      show value for N(c..;d..)\n");
		printf(
"N c .. d .. nn   set N(c..;d..)=nn\n");
		printf(
"E r c .. d ..    print equation corresponding to coefficient of\n");
		printf(
"                 alt: r a range in the form (n1 n2 .. lo-hi ..)\n");
		printf(
"                 exp(..y1)..exp(..yr) prod yk^dk in relation r\n");
		printf(
"W                resume commands from file (current exhaust, if any, \
killed)\n");
		printf(
"X r1 r2 c1 c2    exhaust over range of relations and of c's\n");
		printf(
"                 alt: X (r) c1 c2    (r) a range, e.g. (1 6-9 3 15)\n");
		printf(
"X                restart exhaust after user interrupt\n");
		printf(
"Y                restart exhaust and return control to input file\n");
		printf(
"Z                terminate reading of commands from file (if interrupted)\n");
		printf(
"= [outfile]      display all known N(c,d) (to screen or file)\n");
		printf(
"+                display all N(c,d) with known linear dependences\n");
		printf(
"- [n]            display selected (or all known) linear dependencies\n");
		printf(
"_ [n]            display linear dependencies -- coefs only on RHS\n");
		printf(
":                show program version and current status\n");
		printf(
";anything        comment; no effect\n");
		printf(
"T                display time of day\n");
		printf(
"@                clear overflow flag (if set)\n");
		printf(
"<                break to user input (from a file)\n");
		printf(
"# c ..           set base for positive cone of effective curve classes\n");
		printf(
"*                graphical display\n");
		printf(
"?                display this help screen\n");
		printf(
"Q                quit\n");
	    }
	}
	if (readfile) {
	    if (ov)
		printf("Overflow!\n");
	} else {
	    printf(ov?"OV!>":">");
	}
	nsig = 0;
	signal(SIGINT, userinputsig);
    }
    printf("\n");
    if (readfile)
	return 0;
    else
	exit(0);
}

int main(int argc, char *argv[])
{
    mainloop(1, 0, stdin);
    return 0;
}

