/*
 * Fixed-precision arithmetic
 *  compile with -DPREC=nn to specify precision
 *  requires long to have size at least 4 and short to have size at least 2
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "fixed.h"
#define MASK1 0xffff
#define MASK1BITS 16
#define MASK2 0xff
#define MASK2BITS 8
#define BIGMINUS 0x80000000
#define max(x,y) ((x) > (y) ? (x) : (y))

int ov;

/* set z equal to zero */
void zzero(znumber *z)
{
    z->sign = z->len = 0;
}

/* returns sign of z (+1, 0, or -1) */
int zsgn(znumber *z)
{
    if (z->len == 0)
	return 0;
    else
	return z->sign ? -1 : 1;
}

void zchs(znumber *z)
{
    if (z->len > 0)
	z->sign ^= 1;
}

/* force z to be a legal integer */
void znorm(znumber *z)
{
    if (z->len < 0)
	z->len = 0;
    if (z->len > PREC)
	z->len = PREC;
    while (z->len > 0 && z->s[z->len-1] == 0)
	z->len--;
    if (z->len == 0)
	z->sign = 0;
    z->sign = z->sign ? 1 : 0;
}

/* set z equal to the long value x */
void zset(znumber *z, long x)
{
    if (x == 0) {
	z->sign = z->len = 0;
    } else if (x > 0) {
	z->sign = 0;
	z->s[0] = x&MASK1;
	if (x <= MASK1) {
	    z->len = 1;
	} else {
	    z->s[1] = x>>16;
	    z->len = 2;
	}
    } else if (x == BIGMINUS) {
	z->sign = 1;
	z->len = 2;
	z->s[0] = 0;
	z->s[1] = 0x8000;
    } else {
	z->sign = 1;
	z->s[0] = (-x)&MASK1;
	if (-x <= MASK1) {
	    z->len = 1;
	} else {
	    z->s[1] = (-x)>>16;
	    z->len = 2;
	}
    }
}

void zadd(znumber *x, znumber *y, znumber *z)
{
    znumber a;
    int c, i, m, s;
    long l;

    if (x->len == 0) {
	zcopy(y, z);
    } else if (y->len == 0) {
	zcopy(x, z);
    } else {
	if (x->sign == y->sign) {
	    m = max(x->len, y->len);
	    a.sign = x->sign;
	    c = 0;
	    if (x->len >= y->len) {
		for (i=0; i<y->len; i++) {
		    l = x->s[i] + y->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
		for (; i<x->len; i++) {
		    l = x->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
	    } else {
		for (i=0; i<x->len; i++) {
		    l = x->s[i] + y->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
		for (; i<y->len; i++) {
		    l = y->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
	    }
	    if (c) {
		if (m == PREC) {
		    ov = 1;
		    a.len = m;
		    znorm(&a);
		} else {
		    a.s[m] = 1;
		    a.len = m+1;
		}
	    } else {
		a.len = m;
	    }
	    zcopy(&a, z);
	} else {
	    s = zglem(x, y);
	    if (s == 0) {
		zzero(z);
	    } else {
		m = max(x->len, y->len);
		a.sign = (s==-1);
		c = 0;
		if (a.sign ^ x->sign) { /* |z| = |y| - |x| */
		    for (i=0; i<x->len; i++) {
			l = y->s[i] - x->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		    for (; i<y->len; i++) {
			l = y->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		} else { /* |z| = |x| - |y| */
		    for (i=0; i<y->len; i++) {
			l = x->s[i] - y->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		    for (; i<x->len; i++) {
			l = x->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		}
		while (a.s[m-1] == 0)
		    m--;
		a.len = m;
		zcopy(&a, z);
	    }
	}
    }
}

void zsub(znumber *x, znumber *y, znumber *z)
{
    znumber a;
    int c, i, m, s;
    long l;

    if (x->len == 0) {
	zcopy(y, z);
	zchs(z);
    } else if (y->len == 0) {
	zcopy(x, z);
    } else {
	if (x->sign != y->sign) {
	    m = max(x->len, y->len);
	    a.sign = x->sign;
	    c = 0;
	    if (x->len >= y->len) {
		for (i=0; i<y->len; i++) {
		    l = x->s[i] + y->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
		for (; i<x->len; i++) {
		    l = x->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
	    } else {
		for (i=0; i<x->len; i++) {
		    l = x->s[i] + y->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
		for (; i<y->len; i++) {
		    l = y->s[i] + c;
		    c = (l > MASK1);
		    a.s[i] = l&MASK1;
		}
	    }
	    if (c) {
		if (m == PREC) {
		    ov = 1;
		    a.len = m;
		    znorm(&a);
		} else {
		    a.s[m] = 1;
		    a.len = m+1;
		}
	    } else {
		a.len = m;
	    }
	    zcopy(&a, z);
	} else {
	    s = zgle(x, y);
	    if (s == 0) {
		zzero(z);
	    } else {
		m = max(x->len, y->len);
		a.sign = (s==-1);
		c = 0;
		if (a.sign ^ x->sign) { /* |z| = |y| - |x| */
		    for (i=0; i<x->len; i++) {
			l = y->s[i] - x->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		    for (; i<y->len; i++) {
			l = y->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		} else { /* |z| = |x| - |y| */
		    for (i=0; i<y->len; i++) {
			l = x->s[i] - y->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		    for (; i<x->len; i++) {
			l = x->s[i] - c;
			c = (l < 0);
			a.s[i] = l&MASK1;
		    }
		}
		while (a.s[m-1] == 0)
		    m--;
		a.len = m;
		zcopy(&a, z);
	    }
	}
    }
}

/*
 * multiply x by y
 */
void zmul(znumber *x, short y, znumber *z)
{
    znumber a;
    int i;
    unsigned long c, l;
    unsigned short yy;

    if (y == 0 || x->len == 0) {
	zzero(z);
    } else {
	if (y < 0) {
	    a.sign = x->sign^1;
	    yy = -y;
	} else {
	    a.sign = x->sign;
	    yy = y;
	}
	c = 0;
	for (i=0; i<x->len; i++) {
	    l = x->s[i]*yy + c;
	    c = l >> MASK1BITS;
	    a.s[i] = l&MASK1;
	}
	if (c) {
	    if (x->len == PREC) {
		ov = 1;
		a.len = x->len;
		znorm(&a);
	    } else {
		a.s[x->len] = c;
		a.len = x->len+1;
	    }
	} else {
	    a.len = x->len;
	}
	zcopy(&a, z);
    }
}

/*
 * multiply x by the unsigned quantity y
 */
void zmulu(znumber *x, unsigned short y, znumber *z)
{
    znumber a;
    int i;
    unsigned long c, l;

    if (y == 0 || x->len == 0) {
	zzero(z);
    } else {
	a.sign = x->sign;
	c = 0;
	for (i=0; i<x->len; i++) {
	    l = x->s[i]*y + c;
	    c = l >> MASK1BITS;
	    a.s[i] = l&MASK1;
	}
	if (c) {
	    if (x->len == PREC) {
		ov = 1;
		a.len = x->len;
		znorm(&a);
	    } else {
		a.s[x->len] = c;
		a.len = x->len+1;
	    }
	} else {
	    a.len = x->len;
	}
	zcopy(&a, z);
    }
}

/*
 * multiply x by y
 */
void zzmul(znumber *x, znumber *y, znumber *z)
{
    znumber a, b;
    int i, j;

    if (x->len == 0 || y->len == 0) {
	zzero(z);
    } else {
	if (y->len <= x->len) {
	    zmulu(x, y->s[0], &a);
	    for (i=1; i<y->len; i++) {
		if (y->s[i] != 0) {
		    zmulu(x, y->s[i], &b);
		    if (b.len + i > PREC) {
			for (j=PREC-1; j>=i; j--)
			    b.s[j] = b.s[j-i];
			b.len = PREC;
			ov = 1;
		    } else {
			for (j=b.len + i - 1; j>=i; j--)
			    b.s[j] = b.s[j-i];
			b.len += i;
		    }
		    for (j=0; j<i; j++)
			b.s[j] = 0;
		    zadd(&b, &a, &a);
		}
	    }
	} else {
	    zmulu(y, x->s[0], &a);
	    for (i=1; i<x->len; i++) {
		if (x->s[i] != 0) {
		    zmulu(y, x->s[i], &b);
		    if (b.len + i > PREC) {
			for (j=PREC-1; j>=i; j--)
			    b.s[j] = b.s[j-i];
			b.len = PREC;
			ov = 1;
		    } else {
			for (j=b.len + i - 1; j>=i; j--)
			    b.s[j] = b.s[j-i];
			b.len += i;
		    }
		    for (j=0; j<i; j++)
			b.s[j] = 0;
		    zadd(&b, &a, &a);
		}
	    }
	}
	a.sign = x->sign^y->sign;
	zcopy(&a, z);
    }
}

/*
 * divide x by y (result is z remainder r)
 */
void zdiv(znumber *x, short y, znumber *z, int *r)
{
    znumber a;
    int i, m;
    unsigned long c;
    unsigned short yy;

    if (y == 0) {
	*r = 0;
	ov = 1;
	zzero(z);
    } else if (x->len == 0) {
	*r = 0;
	zzero(z);
    } else {
	if (y < 0) {
	    a.sign = x->sign^1;
	    yy = -y;
	} else {
	    a.sign = x->sign;
	    yy = y;
	}
	c = 0;
	for (i=x->len-1; i>=0; i--) {
	    c = (c<<MASK1BITS) | x->s[i];
	    a.s[i] = c/yy;
	    c %= yy;
	}
	*r = c;
	if (x->sign)
	    *r = -*r;
	m = x->len;
	a.len = a.s[m-1] == 0 ? m-1 : m;
	if (a.len == 0)
	    a.sign = 0;
	zcopy(&a, z);
    }
}

/*
 * divide (signed) x by (unsigned) y (result is z remainder r)
 */
void zdivu(znumber *x, unsigned short y, znumber *z, int *r)
{
    znumber a;
    int i, m;
    unsigned long c;

    if (y == 0) {
	*r = 0;
	ov = 1;
	zzero(z);
    } else if (x->len == 0) {
	*r = 0;
	zzero(z);
    } else {
	a.sign = x->sign;
	c = 0;
	for (i=x->len-1; i>=0; i--) {
	    c = (c<<MASK1BITS) | x->s[i];
	    a.s[i] = c/y;
	    c %= y;
	}
	*r = c;
	if (x->sign)
	    *r = -*r;
	m = x->len;
	a.len = a.s[m-1] == 0 ? m-1 : m;
	if (a.len == 0)
	    a.sign = 0;
	zcopy(&a, z);
    }
}

/*
 * divide x by y (result is z remainder r)
 */
void zzdiv(znumber *x, znumber *y, znumber *z, znumber *r)
{
    znumber a, yy, p, buf;
    int i, j, k, l;
    unsigned long c, d;
    unsigned short v;

    if (y->len == 0) {
	ov = 1;
	zzero(z);
	zzero(r);
    } else if (x->len == 0) {
	zzero(z);
	zzero(r);
    } else if (y->len == 1) {
	v = y->s[0];
	zdivu(x, v, z, &i);
	if (y->sign)
	    zchs(z);
	zset(r, i);
    } else if (y->len == PREC) {
	ov = 1;
	zzero(z);
	zzero(r);
    } else if (x->len < y->len) {
	zzero(z);
	zcopy(x, r);
    } else {
	zcopy(y, &yy);
	yy.sign = 0;
	k = yy.len-1;
	l = x->len-1;
	c = x->s[l];
	buf.sign = 0;
	buf.len = k+1;
	for (i=0; i<=k; i++)
	    buf.s[i] = x->s[l-k+i];
	v = yy.s[k];
	if (v <= MASK2) {
	    v = (v<<MASK2BITS)|(yy.s[k-1]>>MASK2BITS);
	    for (i=0; i<=l-k; i++) {
		d=((c<<MASK2BITS)|MASK2)/v;
		if (d > MASK1)
		    d = MASK1;
		zmulu(&yy, d, &p);
/* {int nloop=0; */
		while (zgle(&p, &buf) > 0) {
/* nloop++; */
		    d--;
		    zsub(&p, &yy, &p);
		}
/* printf("times through loop: %d\n", nloop);} */
		z->s[l-k-i] = d;
		zsub(&buf, &p, &buf);
		if (buf.len >= k+2) {
		    fprintf(stderr, "error: len >= %d\n", k+2);
		    zprintv(&buf);
		}
		if (i < l-k) {
		    for (j=buf.len-1; j>=0; j--)
			buf.s[j+1] = buf.s[j];
		    buf.s[0] = x->s[l-k-i-1];
		    buf.len++;
		    znorm(&buf);
		    c = ((buf.len >= k+2 ? buf.s[k+1] : 0)<<MASK1BITS) |
					(buf.len >= k+1 ? buf.s[k] : 0);
		} else {
		    zcopy(&buf, r);
		}
	    }
	} else {
	    for (i=0; i<=l-k; i++) {
		d = c/v;
		if (d > MASK1)
		    d = MASK1;
		zmulu(&yy, d, &p);
/* {int nloop=0; */
		while (zgle(&p, &buf) > 0) {
/* nloop++; */
		    d--;
		    zsub(&p, &yy, &p);
		}
/* printf("times through loop: %d\n", nloop);} */
		z->s[l-k-i] = d;
		zsub(&buf, &p, &buf);
		if (buf.len >= k+2) {
		    fprintf(stderr, "error: len >= %d\n", k+2);
		    zprintv(&buf);
		}
		if (i < l-k) {
		    for (j=buf.len-1; j>=0; j--)
			buf.s[j+1] = buf.s[j];
		    buf.s[0] = x->s[l-k-i-1];
		    buf.len++;
		    znorm(&buf);
		    c = ((buf.len >= k+2 ? buf.s[k+1] : 0)<<MASK1BITS) |
					(buf.len >= k+1 ? buf.s[k] : 0);
		} else {
		    zcopy(&buf, r);
		}
	    }
	}
	z->sign = x->sign ^ y->sign;
	z->len = l-k+1;
	znorm(z);
	r->sign = x->sign;
	znorm(r);
    }
/* TEST: check it! */ /* {znumber ch; zzmul(z,y,&ch); zadd(&ch,r,&ch);
if (zcmp(&ch,x)) {printf("division error! %d\n",ov); printf("%s / ",ztodec(x));
printf("%s =? ", ztodec(y)); printf("%s r ", ztodec(z));
printf("%s\n", ztodec(r)); zprintv(x); zprintv(y); zprintv(z); zprintv(r);
exit(1);} } */
}

/*
 * Return +1/0/-1 according to whether x is bigger/equal/smaller than y
 */
int zgle(znumber *x, znumber *y)
{
    int i;

    if (x->sign == y->sign) {
	if (x->len == y->len) {
	    for (i=x->len-1; i>=0; i--)
		if (x->s[i] > y->s[i])
		    return x->sign ? -1 : 1;
		else if (x->s[i] < y->s[i])
		    return x->sign ? 1 : -1;
	    return 0;
	} else {
	    return (x->sign ^ (x->len > y->len)) ? 1 : -1;
	}
    } else {
	return y->sign - x->sign;
    }
}

/*
 * Return +1/0/-1 according to whether x is bigger/equal/smaller than -y
 */
int zglem(znumber *x, znumber *y)
{
    int i;

    if (x->sign != y->sign) {
	if (x->len == y->len) {
	    for (i=x->len-1; i>=0; i--)
		if (x->s[i] > y->s[i])
		    return x->sign ? -1 : 1;
		else if (x->s[i] < y->s[i])
		    return x->sign ? 1 : -1;
	    return 0;
	} else {
	    return (x->sign ^ (x->len > y->len)) ? 1 : -1;
	}
    } else {
	if (x->len == 0 && y->len == 0)
	    return 0;
	return (1^y->sign) - x->sign;
    }
}

void zprintv(znumber *z)
{
    int i;

    printf("( %s ; %4d ; ", z->sign ? "MINUS" : " PLUS", z->len);
    if (z->len == 0)
	printf(" [ZERO] )\n");
    else
	for (i=0; i<z->len; i++)
	    printf("%s%7u%s",
		    i==0 ? "" : "                 ",
		    z->s[i],
		    i==z->len-1 ? " )\n" : "\n");
}

char *ztodec(znumber *z)
{
    static char s[5*PREC + 2];
    char *p;
    const short d = 10;
    znumber a;
    int r;

    s[5*PREC + 1] = '\0';
    if (z->len == 0) {
	s[5*PREC] = '0';
	return &s[5*PREC];
    }
    p = &s[5*PREC];
    zcopy(z, &a);
    a.sign = 0;
    while (a.len) {
	zdiv(&a, d, &a, &r);
	*p-- = r + '0';
    }
    if (z->sign) {
	*p = '-';
	return p;
    } else {
	return p+1;
    }
}

void dectoz(char s[], znumber *z, char **zend)
{
    int sign;
    char *p;
    znumber x, y;

    p = s;
    while (*p != '\0' && isspace(*p))
	p++;
    if (*p == '\0') {
	zzero(z);
	if (zend)
	    *zend = s;
	return;
    }
    sign = 0;
    if (*p == '-' || *p == '+') {
	sign = *p == '-';
	p++;
	while (*p != '\0' && isspace(*p))
	    p++;
	if (*p == '\0') {
	    zzero(z);
	    if (zend)
		*zend = s;
	    return;
	}
    }
    if (!isdigit(*p)) {
	zzero(z);
	if (zend)
	    *zend = s;
	return;
    }
    zset(z, *p - '0');
    p++;
    while (*p != '\0' && isdigit(*p)) {
	zmul(z, 10, &y);
	zset(&x, *p - '0');
	zadd(&y, &x, z);
	p++;
    }
    if (sign)
	zchs(z);
    if (zend)
	*zend = p;
}

double ztof(znumber *z)
{
    double x;
    int i;

    x = 0.0;
    for (i=0; i<z->len; i++)
	x += ldexp((double)z->s[i], i*MASK1BITS);
    return z->sign ? -x : x;
}

void qzero(qnumber *z)
{
    z->n.sign = z->n.len = 0;
    z->d.sign = 0;
    z->d.len = 1;
    z->d.s[0] = 1;
}

void qreduce(qnumber *z)
{
    znumber a, b, g, rem;
    void mgcd(znumber *, znumber *, znumber *);

    zcopy(&z->n, &a);
    zcopy(&z->d, &b);
    mgcd(&a, &b, &g);
    zzdiv(&a, &g, &z->n, &rem);
    zzdiv(&b, &g, &z->d, &rem);
}

void ztoq(znumber *x, qnumber *z)
{
    zcopy(x, &z->n);
    z->d.sign = 0;
    z->d.len = 1;
    z->d.s[0] = 1;
}

void zztoq(znumber *x, znumber *y, qnumber *z)
{
    zcopy(x, &z->n);
    zcopy(y, &z->d);
    if (z->d.len == 0) {
	z->d.sign = 0;
	z->d.len = 1;
	z->d.s[0] = 1;
	ov = 1;
    }
    if (z->d.sign == 1) {
	zchs(&z->n);
	zchs(&z->d);
    }
    qreduce(z);
}

void qnorm(qnumber *z)
{
    znorm(&z->n);
    znorm(&z->d);
    if (z->d.len == 0) {
	z->d.sign = 0;
	z->d.len = 1;
	z->d.s[0] = 1;
    }
    if (z->d.sign == 1) {
	zchs(&z->n);
	zchs(&z->d);
    }
    qreduce(z);
}

void qset(qnumber *z, long x, long y)
{
    zset(&z->n, x);
    zset(&z->d, y);
    qnorm(z);
}

void qadd(qnumber *x, qnumber *y, qnumber *z)
{
    znumber a, b;

    zzmul(&x->d, &y->d, &z->d);
    zzmul(&x->n, &y->d, &a);
    zzmul(&x->d, &y->n, &b);
    zadd(&a, &b, &z->n);
    qreduce(z);
}

void qsub(qnumber *x, qnumber *y, qnumber *z)
{
    znumber a, b;

    zzmul(&x->d, &y->d, &z->d);
    zzmul(&x->n, &y->d, &a);
    zzmul(&x->d, &y->n, &b);
    zsub(&a, &b, &z->n);
    qreduce(z);
}

void qmul(qnumber *x, short y, qnumber *z)
{
    zmul(&x->n, y, &z->n);
    zcopy(&x->d, &z->d);
    qreduce(z);
}

void qmulu(qnumber *x, unsigned short y, qnumber *z)
{
    zmulu(&x->n, y, &z->n);
    zcopy(&x->d, &z->d);
    qreduce(z);
}

void qqmul(qnumber *x, qnumber *y, qnumber *z)
{
    zzmul(&x->n, &y->n, &z->n);
    zzmul(&x->d, &y->d, &z->d);
    qreduce(z);
}

void qdiv(qnumber *x, short y, qnumber *z)
{
    zmul(&x->d, y, &z->d);
    zcopy(&x->n, &z->n);
    if (z->d.len == 0) {
	z->d.sign = 0;
	z->d.len = 1;
	z->d.s[0] = 1;
	ov = 1;
    }
    if (z->d.sign == 1) {
	zchs(&z->n);
	zchs(&z->d);
    }
    qreduce(z);
}

void qdivu(qnumber *x, unsigned short y, qnumber *z)
{
    zmulu(&x->d, y, &z->d);
    zcopy(&x->n, &z->n);
    if (z->d.len == 0) {
	z->d.sign = 0;
	z->d.len = 1;
	z->d.s[0] = 1;
	ov = 1;
    }
    if (z->d.sign == 1) {
	zchs(&z->n);
	zchs(&z->d);
    }
    qreduce(z);
}

void qqdiv(qnumber *x, qnumber *y, qnumber *z)
{
    zzmul(&x->n, &y->d, &z->n);
    zzmul(&x->d, &y->n, &z->d);
    if (z->d.len == 0) {
	z->d.sign = 0;
	z->d.len = 1;
	z->d.s[0] = 1;
	ov = 1;
    }
    if (z->d.sign == 1) {
	zchs(&z->n);
	zchs(&z->d);
    }
    qreduce(z);
}

void qinv(qnumber *y, qnumber *z)
{
    zcopy(&y->n, &z->d);
    zcopy(&y->d, &z->n);
    if (z->d.len == 0) {
	z->d.sign = 0;
	z->d.len = 1;
	z->d.s[0] = 1;
	ov = 1;
    }
    if (z->d.sign == 1) {
	zchs(&z->n);
	zchs(&z->d);
    }
}

int qgle(qnumber *x, qnumber *y)
{
    znumber a, b;

    zzmul(&x->n, &y->d, &a);
    zzmul(&y->n, &x->d, &b);
    return zgle(&a, &b);
}

int qglem(qnumber *x, qnumber *y)
{
    znumber a, b;

    zzmul(&x->n, &y->d, &a);
    zzmul(&y->n, &x->d, &b);
    return zglem(&a, &b);
}

void qprintv(qnumber *z)
{
    printf("NUMER\n");
    zprintv(&z->n);
    printf("DENOM\n");
    zprintv(&z->d);
}

char *qtodec(qnumber *z)
{
    static char n[10*PREC + 6];
    char d[5*PREC + 5];

    if (z->d.sign == 0 && z->d.len == 1 && z->d.s[0] == 1)
	return ztodec(&z->n);
    strcpy(n, ztodec(&z->n));
    d[0] = ' ';
    d[1] = '/';
    d[2] = ' ';
    strcpy(d+3, ztodec(&z->d));
    return strcat(n, d);
}

/* valid formats:   3/4   -3/4   -(3/4)   (-3/4) */
void dectoq(char s[], qnumber *q, char **qend)
{
    int sign, b, c;
    char *p;
    znumber x, y, n, d;

    p = s;
    b = 0;
    while (*p != '\0' && (isspace(*p) || (b == 0 && (b+=(*p=='(')) == 1)))
	p++;
    if (*p == '\0') {
	qzero(q);
	if (qend)
	    *qend = s;
	return;
    }
    sign = 0;
    if (*p == '-' || *p == '+') {
	sign = *p == '-';
	p++;
	while (*p != '\0' && (isspace(*p) || (b == 0 && (b+=(*p=='(')) == 1)))
	    p++;
	if (*p == '\0') {
	    qzero(q);
	    if (qend)
		*qend = s;
	    return;
	}
    }
    if (!isdigit(*p)) {
	qzero(q);
	if (qend)
	    *qend = s;
	return;
    }
/* numerator */
    zset(&n, *p - '0');
    p++;
    while (*p != '\0' && isdigit(*p)) {
	zmul(&n, 10, &y);
	zset(&x, *p - '0');
	zadd(&y, &x, &n);
	p++;
    }
    if (sign)
	zchs(&n);
    c = b;
    while (*p != '\0' && (isspace(*p) || (c == 1 && (c-=(*p==')')) == 0)))
	p++;
    if (b == 1 && c == 0) {         /* string is  ( ... )   */
	zset(&d, 1);
	zztoq(&n, &d, q);
	if (qend)
	    *qend = p;
	return;
    }
/* denominator is optional, but parentheses must be closed */
    if (*p != '/') {
	if (b == 1) {
	    qzero(q);
	    if (qend)
		*qend = s;
	} else {
	    zset(&d, 1);
	    zztoq(&n, &d, q);
	    if (qend)
		*qend = p;
	}
	return;
    }
    p++;
    while (*p != '\0' && isspace(*p))
	p++;
/* denominator */
    if (!isdigit(*p)) {
	qzero(q);
	if (qend)
	    *qend = s;
	return;
    }
    zset(&d, *p - '0');
    p++;
    while (*p != '\0' && isdigit(*p)) {
	zmul(&d, 10, &y);
	zset(&x, *p - '0');
	zadd(&y, &x, &d);
	p++;
    }
/* force closing parenthesis (if needed) */
    while (*p != '\0' && (isspace(*p) || (b == 1 && (b-=(*p==')')) == 0)))
	p++;
    if (b == 0) {
	zztoq(&n, &d, q);
	if (qend)
	    *qend = p;
	return;
    } else {
	qzero(q);
	if (qend)
	    *qend = s;
	return;
    }
}


double qtof(qnumber *z)
{
    return ztof(&z->n)/ztof(&z->d);
}

