/*
 * for input/output of polynomials whose coefficients are
 * rational numbers
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include "fixed.h"

#define min(x,y) ((x) > (y) ? (y) : (x))

#define MACAULAY 1
#define HUMAN 2
#define HUMANCOMPACT 3
#define HUMANFORREENTRY 4
#define CLEAR_DENOMS 1
#define COMMON_DENOMS 2
#define NO_DENOMS 3

#define MAXDENOM 17
#define SMALL 30

#define NUMLEN(n) ( (n) < 10 ? 1 : \
		    (n) < 100 ? 2 : \
		    (n) < 1000 ? 3 : \
		    (n) < 10000 ? 4 : \
		    (n) < 100000 ? 5 : 6 )

int fgetline(FILE *, char *, int);

void bytecopy(char *a, char *b, int n)
{
    int i;

    for (i=0; i<n; i++)
	b[i] = a[i];
}

/*
 * print a vector representing a (homogeneous) polynomial in nice form
 * inputs:
 *  v: input vector
 *  nterms: number of terms, generally (n+d-1 choose d)
 *  p: characteristic, a prime less than 32767
 *  ptab: table of inverses mod p
 *  nvar: number of variables
 *  expn: table of exponents of the nvar variables for each term
 *      term k represents a_1^expn[k*nvar]*...*a_nvar^expn[k*nvar+nvar-1]
 *  vname: names of the nvar variables
 *  vl: lengths, in characters, of the variable names: vl[i] = strlen(vname[i])
 *  style: style of output HUMAN or MACAULAY
 *         or HUMANCOMPACT (like HUMAN but blank lines are not printed)
 *         or HUMANFORREENTRY (like HUMAN but exponents printed on same line)
 *  denoms: what to do about denominators (which show up as large integers
 *   mod p) -- clear them by scaling, put whole expr over common denominator,
 *   or do nothing
 *  scr: scratch space, of length nterms
 *
 * due to initialization, p should be the same every time this routine is
 * called
 */
void fnicempvect(FILE *whereto, qnumber *v, int nterms, int nvar, int *expn,
char **vname, int *vl, int style, int denoms, int order)
{
    int i, ii, j, k, l, t, t0, c1, c2, p2, s, nsp, nat;
    int firstout, closeparen, dstrlen, havedenom;
    char ss1[180], ss2[180], dstr[50], *s1, *s2, *p;
    znumber denom, x, y, z, one;
    qnumber denomq, oneq, c, cz;
    void mgcd(znumber *, znumber *, znumber *);

    if (style == HUMANFORREENTRY) {
	s1 = ss1;
	s2 = ss1;
    } else {
	s1 = ss1;
	s2 = ss2;
    }
    zset(&one, 1);
    qset(&oneq, 1, 1);
/* clear denominators */
    if (denoms != NO_DENOMS) {
	zset(&denom, 1);
	for (i=0; i<nterms; i++) {
	    mgcd(&denom, &v[i].d, &z);
	    zzdiv(&denom, &z, &y, &x);
	    zzmul(&v[i].d, &y, &denom);
	}
	havedenom = zgle(&denom, &one) ? 1 : 0;
	ztoq(&denom, &denomq);
    } else {
	havedenom = 0;
    }
    if (denoms == COMMON_DENOMS && havedenom) {
	if (style == HUMAN || style == HUMANCOMPACT ||
					    style == HUMANFORREENTRY)
	    sprintf(dstr, " ) / %s", ztodec(&denom));
	else
	    sprintf(dstr, "/%d", ztodec(&denom));
	dstrlen = strlen(dstr);
    }
    t = 0;
    firstout = 1;
    closeparen = 0;
    for (i=nterms-1; i>=-2; i--) {
	if (i == -2 || (i == -1 && closeparen) ||
			(i >= 0 && qsgn(&v[order ? i : nterms - 1 - i]))) {
	    t0 = t;
	    if (i >= 0) {
		ii = order ? i : nterms - 1 - i;
		qcopy(&v[ii], &c);
		if (qsgn(&v[ii]) < 0) {
		    s = 1;
		    qchs(&c);
		} else {
		    s = 0;
		}
		if (havedenom) {
		    qqmul(&c, &denomq, &cz);
		    qcopy(&cz, &c);
		}
/* coefficient */
		if (style == HUMAN || style == HUMANCOMPACT ||
					    style == HUMANFORREENTRY) {
		    s1[t] = ' ';
		    s2[t] = ' ';
		    t++;
		}
		if (firstout && denoms == COMMON_DENOMS && havedenom &&
				(style == HUMAN || style == HUMANCOMPACT ||
				style == HUMANFORREENTRY)) {
		    s1[t] = ' ';
		    s2[t] = '(';
		    s1[t+1] = ' ';
		    s2[t+1] = ' ';
		    t += 2;
		    closeparen = 1;
		}
		if (!firstout || s) {
		    s2[t] = s ? '-' : '+';
		    if (style == HUMAN || style == HUMANCOMPACT ||
				style == HUMANFORREENTRY) {
			s2[t+1] = ' ';
			if (style != HUMANFORREENTRY) {
			    s1[t] = ' ';
			    s1[t+1] = ' ';
			}
			t += 2;
		    } else {
			t++;
		    }
		}
		k = 0;
		for (j=0; j<nvar && !k; j++)
		    k = expn[ii*nvar+j];
		if (qgle(&c, &oneq) || !k || (denoms == COMMON_DENOMS &&
				    havedenom && style == MACAULAY)) {
		    l = zgle(&c.d, &one) ? 1 : 0;
		    k = strlen(p=qtodec(&c)) + 2*l;
		    if (style == HUMAN || style == HUMANCOMPACT) {
			for (j=0; j<k; j++)
			    s1[t+j] = ' ';
		    }
		    for (j=0; j<k-2*l; j++) {
			s2[t+j+l] = p[j];
		    }
		    if (l) {
			s2[t] = '(';
			s2[t+k-1] = ')';
		    }
		    t += k;
		    if (style == HUMAN || style == HUMANCOMPACT ||
					style == HUMANFORREENTRY) {
			s1[t] = ' ';
			s2[t] = ' ';
			t++;
		    } else if (denoms == COMMON_DENOMS && havedenom) {
			bytecopy(dstr, &s2[t], dstrlen);
			t += dstrlen;
		    }
		}
/* variable name */
		for (j=0; j<nvar; j++) {
		    if (expn[ii*nvar+j]) {
			bytecopy(vname[j], &s2[t], vl[j]);
			if (style == HUMAN || style == HUMANCOMPACT)
			    for (k=0; k<vl[j]; k++)
				s1[t+k] = ' ';
			t += vl[j];
			if (style == HUMAN || style == HUMANCOMPACT) {
			    s2[t] = ' ';
			    if (expn[ii*nvar+j] > 1)
				s1[t] = '0' + expn[ii*nvar+j];
			    else
				s1[t] = ' ';
			    t++;
			} else if (style == HUMANFORREENTRY) {
			    if (expn[ii*nvar+j] > 1) {
				s1[t] = '^';
				t++;
				s1[t] = '0' + expn[ii*nvar+j];
				t++;
			    }
			    s1[t] = ' ';
			    t++;
			} else {
			    if (expn[ii*nvar+j] > 1) {
				s2[t] = '0' + expn[ii*nvar+j];
				t++;
			    }
			}
		    }
		}
		firstout = 0;
	    } else if (i == -1) {
		for (k=0; k<dstrlen; k++)
		    s1[t+k] = ' ';
		bytecopy(dstr, &s2[t], dstrlen);
		t += dstrlen;
	    }
/* print out as far as t0 */
	    if (i == -2 || t >= ((style == HUMAN || style == HUMANCOMPACT) ?
							    79 : 77)) {
		j = t0;
		while (j > 0 && s1[j-1] == ' ')
		    j--;
		c1 = s1[j];
		s1[j] = '\0';
		if (style == HUMANFORREENTRY) {
		    k = j;
		} else {
		    k = t0;
		    while (k > 0 && s2[k-1] == ' ')
			k--;
		    c2 = s2[k];
		    s2[k] = '\0';
		}
		if (j > 0 || k > 0) {
		    if (style == HUMAN) {
			fputs(s1, whereto);
			putc('\n', whereto);
			fputs(s2, whereto);
			putc('\n', whereto);
		    } else if (style == HUMANCOMPACT) {
			if (*s1 != '\0') {
			    fputs(s1, whereto);
			    putc('\n', whereto);
			}
			if (*s2 != '\0') {
			    fputs(s2, whereto);
			    putc('\n', whereto);
			}
		    } else if (style == HUMANFORREENTRY) {
			fputs(s2, whereto);
			if (i >= 0) {
			    putc(' ', whereto);
			    putc('\\', whereto);
			}
			putc('\n', whereto);
		    } else {
			fputs(s2, whereto);
			if (i >= 0) {
			    putc(' ', whereto);
			    putc('\\', whereto);
			}
			putc('\n', whereto);
		    }
		}
		if (style == HUMAN || style == HUMANCOMPACT ||
						style == HUMANFORREENTRY)
		    s1[j] = c1;
		if (style != HUMANFORREENTRY)
		    s2[k] = c2;
		if (i >= -1) {
		    if (style == HUMAN || style == HUMANCOMPACT ||
						style == HUMANFORREENTRY)
			bytecopy(&s1[t0], &s1[6], t-t0);
		    if (style != HUMANFORREENTRY)
			bytecopy(&s2[t0], &s2[6], t-t0);
		    t = t - t0 + 6;
/* indent subsequent lines */
		    for (j=0; j<6; j++) {
			if (style == HUMAN || style == HUMANCOMPACT ||
						style == HUMANFORREENTRY)
			    s1[j] = ' ';
			s2[j] = ' ';
		    }
		}
	    }
	}
    }
    if (firstout) {
	if (style == HUMAN || style == HUMANCOMPACT ||
						style == HUMANFORREENTRY) {
	    fprintf(whereto, "\n 0\n");
	} else {
	    fprintf(whereto, "0\n");
	}
    }
}

void nicempvect(qnumber *v, int nterms, int nvar, int *expn, char **vname,
int *vl, int style, int denoms, int order)
{
    fnicempvect(stdout, v, nterms, nvar, expn, vname, vl, style,
						denoms, order);
}

/*
 * compare the n characters beginning at x with the list of nvar names
 * in vname and return a 1 if a match or partial match is found, with
 * which set to which one and c1 set to last character in matched part
 * of input string in the event of a perfect match; return 0 otherwise
 */
int isaname(char *x, int n, int nvar, char **vname, int *vl, int *which,
char **cend)
{
    int i, j, k, rv;

/* DEBUG printf("%d chars\n", n); */
/* DEBUG printf("is "); */
/* DEBUG for (i=0; i<n; i++) putchar(x[i]); */
/* DEBUG printf("a name?\n"); */
/* DEBUG printf("%d vars\n",  nvar); */
/* DEBUG for (i=0; i<nvar; i++) printf("%d  %s\n",   vl[i], vname[i]); */
    rv = 0;
    for (i=0; i<nvar; i++) {
	if (vl[i] >= n) {
	    k = 1;
	    for (j=0; j<n && k; j++)
		k = (x[j] == vname[i][j]);
	    if (k) {
		rv = 1;
		if (vl[i] == n) {
		    *which = i;
		    *cend = x + n - 1;
/* DEBUG printf("yes   %d\n", i); */
		    return 1;
		}
	    }
	}
    }
    return rv;
}

/*
 * Scans from stdin a polynomial in MACAULAY format, returning
 * 0 for success, 1 for error (e.g. encountered bad format, or a term
 * not in the expn list), or -1 for EOF.
 * On error, a pointer to the offending line is returned in oline
 * if non-NULL (use strdup if you want to keep it around)
 *
 * parameters as in nicepvect, except that for this routine, expect
 * scr to be an array of length nvar
 *
 * examples of valid entries:       vname[] = {"x", "y", "z"}
 *   x2-y
 *   xx - y
 *   -(+2x2-y-x2)/2
 *   x^2 - (1/2)y - 1/2y
 *   x2 - \
 *     y
 *   x2 \
 *     - y
 *
 * vname[] = {"u", "u1", "u2" }
 *   uuu1u22
 *   u 2u1u22
 *   u^2u1u2^2
 *
 * spaces are ignored, expect they cannot occur in the middle of a variable
 * name; optional ^ for exponent; \ for continuation may come between terms
 */
int fscanmvect(FILE *wherefrom, qnumber *v0, int nterms, int nvar, int *expn,
char **vname, int *vl, int *scr, char **oline)
{
    int d, i, j, s, si, gs, dn, more, ex, iparen, firstline;
    char *c, *c0, *c1;
    qnumber cf, z;
    static char li[1000];

    iparen = 0;
    firstline = 1;
    gs = 0;
    for (j=0; j<nterms; j++)
	qzero(&v0[j]);
    si = 2;
    while (1) {
	if (fgetline(wherefrom, li, 999) == EOF)
	    return EOF;
	c = li;
	while (*c != '\0') {
	    if (*c == '*')
		*c = ' ';
	    c++;
	}
/* DEBUG printf("line %s\n", li); */
	j = 1;
	c = li;
	while (j && *c != '\0') {
	    j = isspace(*c);
	    c++;
	}
	if (j)          /* blank line */
	    continue;
	c--;
	if (firstline) {
	    if (*c == '+' || *c == '-') {
		if (*c == '-')
		    gs = 1;
		c0 = c + 1;
		while (*c0 != '\0' && isspace(*c0))
		    c0++;
	    } else {
		c0 = c;
	    }
	    if (*c0 == '(') {
		dectoq(c0, &z, &c1);
				/* distinguish (3x+5y+..)/7 from (3/2)x + .. */
		if (c1 == c0) {
		    iparen = 1;
		} else {        /* also distinguish (3)/7 from (3/2)x + .. */
		    while (*c1 != '\0' && isspace(*c1))
			c1++;
		    if (*c1 == '/')
			iparen = 1;
		}
		c = c0+iparen;
	    }
	}
	while (1) {
	    while (*c != '\0' && isspace(*c))
		c++;
	    d = *c;
	    if (d == '\0') {
		if (iparen)         /* forgot to close parentheses */
		    { if (oline) *oline = li; return 1; }
		return 0;
	    }
	    if (d == '\\')
		break;
	    if (d == ')') {         /* allowable:  ... )   or    ... ) / d */
		if (!iparen)
		    { if (oline) *oline = li; return 1; }
		c++;
		while (*c != '\0' && isspace(*c))
		    c++;
		if (*c != '/' && *c != '\0')
		    { if (oline) *oline = li; return 1; }
		if (*c == '/') {
		    c++;
		    dectoq(c, &cf, &c0);
		    if (c0 == c)
			{ if (oline) *oline = li; return 1; }
		    c = c0;
		    while (*c != '\0' && isspace(*c))
			c++;
		    if (*c != '\0')
			{ if (oline) *oline = li; return 1; }
		    if (gs)
			qchs(&cf);
		    for (j=0; j<nterms; j++) {
			if (qsgn(&v0[j])) {
			    qqdiv(&v0[j], &cf, &z);
			    qcopy(&z, &v0[j]);
			}
		    }
		    return 0;
		} else {
		    if (gs) {
			for (j=0; j<nterms; j++)
			    qchs(&v0[j]);
		    }
		    return 0;
		}
	    } else if (d == '+' || d == '-') {
		if (d == '+')
		    si = 1;
		else
		    si = -1;
		c++;
		while (*c != '\0' && isspace(*c))
		    c++;
		d = *c;
		if (d == '\0')
		    { if (oline) *oline = li;
					/* printf("DEBUG A\n"); */ return 1; }
		if (d == '\\') {    /* allow, e.g.     ...  3x - \ */
		    si *= 2;
		    break;          /*              4y + ...        */
		}
	    } else {
		if (si == 2)        /* sign mandatory except at beginning */
		    si = 1;         /* or if there was sign on prev. line */
		else if (si == -2)
		    si = -1;
		else
		    { if (oline) *oline = li;
					/* printf("DEBUG B\n"); */ return 1; }
	    }
	    dectoq(c, &cf, &c0);
	    if (c0 == c) {
		qset(&cf, 1, 1);
	    } else {
		c = c0;
		d = *c;
		if (d == '\0' || d == '+' || d == '-' || d == '\\' || d == ')')
									    {
			/* search expn for all exponents zero (constant term)*/
		    s = 1;
		    for (j=0; j<nterms && s; j++) {
			s = 0;
			for (i=0; i<nvar && !s; i++)
			    s = expn[j*nvar+i];
		    }
		    if (qsgn(&cf)) {
			if (s)
			    { if (oline) *oline = li; /* printf("DEBUG D\n");*/
								    return 1; }
			if (gs && !iparen) {
			    si = -si;
			    gs = 0;
			}
			if (si == 1) {      /* store coef in constant term */
			    qadd(&v0[j-1], &cf, &z);
			    qcopy(&z, &v0[j-1]);
			} else {
			    qsub(&v0[j-1], &cf, &z);
			    qcopy(&z, &v0[j-1]);
			}
		    }
		    continue;           /* and go on to next term (if any) */
		}
	    }
/* so far: (+ or -) (optional coefficient) variable names with exponents */
/*           we are here ------------------^              */
	    for (i=0; i<nvar; i++)
		scr[i] = 0;
	    while (*c != '\0' && *c != '\\' && *c != '+' && *c != '-' &&
								*c != ')') {
		c0 = c;     /* mark start of variable name */
		i = -1;
		while (*c != '\0' && isaname(c0, c-c0+1, nvar, vname, vl,
								    &i, &c1))
		    c++;
		if (i == -1)
		    { if (oline) *oline = li; /* printf("DEBUG E\n"); */
								    return 1; }
/* DEBUG printf("got variable number %d\n", i); */
		c = c1+1;
		while (*c != '\0' && (isspace(*c) || *c == '^'))
		    c++;
		if (*c == '\0' || !isdigit(*c)) {
/* DEBUG printf("power %d plus 1", i); */
		    scr[i]++;
/* DEBUG printf("     to "); */
/* DEBUG { int H ; for (H=0; H<nvar; H++)  printf("%d ", scr[H]);   }
							    printf("\n"); */
		} else {
		    ex = 0;
		    while (*c != '\0' && (isspace(*c) || isdigit(*c))) {
			if (isdigit(*c))
			    ex = ex*10 + *c - '0';
			c++;
		    }
		    scr[i] += ex;
		}
	    }
	    s = 0;
	    for (i=0; i<nvar && !s; i++)
		s = scr[i];
	    if (!s)         /* constant term must have explicit coefficient */
		{ if (oline) *oline = li; /* printf("DEBUG F\n");*/ return 1; }
	    s = 1;
/* DEBUG for (j=0; j<nterms; j++)   {    for (i=0; i<nvar;  i++)   {
printf("%d ", expn[j*nvar+i]);    }  printf("\n");   } */
	    for (j=0; j<nterms && s; j++) {
		s = 0;
		for (i=0; i<nvar && !s; i++)
		    s = (expn[j*nvar+i] != scr[i]);
	    }
	    if (qsgn(&cf)) {
		if (s)
		    { if (oline) *oline = li; /* printf("DEBUG G\n"); */
							    return 1; }
		if (gs && !iparen) {
		    si = -si;
		    gs = 0;
		}
		if (si == 1) {
		    qadd(&v0[j-1], &cf, &z);
		    qcopy(&z, &v0[j-1]);
		} else {
		    qsub(&v0[j-1], &cf, &z);
		    qcopy(&z, &v0[j-1]);
		}
	    }
	}
	firstline = 0;
    }
}

int scanmvect(qnumber *v0, int nterms, int nvar, int *expn, char **vname,
int *vl, int *scr, char **oline)
{
    return fscanmvect(stdin, v0, nterms, nvar, expn, vname, vl, scr, oline);
}
