#include <stdio.h>
#include "fixed.h"

/*
 * Invert an n x n matrix of rational multi-precision numbers
 * INPUT: n x n matrix a, stored in an n x hsize array of qnumbers
 *        (2n) x n size workspace b of qnumbers
 *        n-long workspace prmi of ints
 * OUTPUT: n x n matrix a^{-1}, stored in a
 *         return value 1 if error (matrix not invertible), 0 if success
 */
int qmatinv(qnumber *a, qnumber *b, int *prmi, int n, int hsize)
{
    int i, j, k, l, z;
    qnumber mu, mi, v, q1, q2;

/* column-reduce matrix a, keeping track of where reducers lie */
    for (i=0; i<n; i++) {
	for (j=0; j<n; j++) {
	    qzero(&b[j+i*n]);
	}
	qset(&b[i+i*n], 1, 1);
    }
    for (j=0; j<n; j++) {
	z = 1;              /* all zero column? */
	for (i=0; i<n; i++) {
	    if (qsgn(&a[j+i*hsize])) {
		z = 0;
		break;
	    }
	}
	if (z) {
	    return 1;
	} else {            /* record i and reduce with a_ij */
	    prmi[i] = j;
	    qinv(&a[j+i*hsize], &mi);
	    qchs(&mi);
	    for (k = j+1; k < n; k++) {
		if (qsgn(&a[k+i*hsize])) {
		    qqmul(&mi, &a[k+i*hsize], &mu);
		    qzero(&a[k+i*hsize]);
		    for (l=i+1; l<n; l++) {
			qqmul(&mu, &a[j+l*hsize], &q1);
			qadd(&a[k+l*hsize], &q1, &q2);
			qcopy(&q2, &a[k+l*hsize]);
		    }
		    for (l=0; l<=j; l++) {
			qqmul(&mu, &b[j+l*n], &q1);
			qadd(&b[k+l*n], &q1, &q2);
			qcopy(&q2, &b[k+l*n]);
		    }
		}
	    }
	}
    }

/*
 * inverse is a^{-1}b
 * a is lower triangular when appropriately permuted
 * b is upper triangular
 */
    for (i=0; i<n; i++) {           /* compute a^{-1} and store in workspace */
	if (qsgn(&a[prmi[i]+i*hsize]) == 0) {
	    fprintf(stderr,
		    "matinvmodp: column reduction gives 0 on diagonal\n");
	    exit(1);
	}
	qinv(&a[prmi[i]+i*hsize], &mi);
	for (j=0; j<i; j++) {
	    qzero(&v);
	    for (k=j; k<i; k++) {
		qqmul(&a[prmi[k]+i*hsize], &b[j+(n+k)*n], &q1);
		qadd(&v, &q1, &q2);
		qcopy(&q2, &v);
	    }
	    qchs(&v);
	    qqmul(&v, &mi, &b[j+(n+i)*n]);
	}
	qcopy(&mi, &b[i+(n+i)*n]);
	for (j=i+1; j<n; j++)
	    qzero(&b[j+(n+i)*n]);
    }
    for (i=0; i<n; i++) {       /* multiply b by a^{-1}, putting result in a */
	for (j=0; j<n; j++) {
	    qzero(&v);
	    for (k=0; k<n; k++) {
		qqmul(&b[prmi[k]+i*n], &b[j+(n+k)*n], &q1);
		qadd(&v, &q1, &q2);
		qcopy(&q2, &v);
	    }
	    qcopy(&v, &a[j+i*hsize]);
	}
    }
    return 0;
}
